from orionpy.orioncore.Filter import _FDUFilter
from .orioncsv import OrionCSV


class CSVOrganisationUnit(OrionCSV):
    # TODO handle labels starting with a 0 !
    def __init__(self, orion, storage_id=""):
        super().__init__(orion)
        self.username_head = 'UserName'
        self.ou_head = 'Organization'
        self.username_i = 0
        self.ou_i = 1
        self.storage_id = storage_id or 'hosted'


    def generate(self, csv_path):
        """Create/Edit a csv file with a summary of all FDU and filtering set for all groups

        :param csv_path: path to the csv files
        """
        print('Generating csv file with a summary of all FDU and filtering values set...')
        # Prepares data and header required
        header = [self.username_head, self.ou_head]
        users = self.orion.users.all()

        storage_resource = self.orion.businesses.get_stats_resource(self.storage_id)
        if storage_resource is None:
            print('No stats resource define')
            return None
        
        filter_id = storage_resource.filter_id
        if filter_id is None:
            print('No filter associated with stats resource')
            return None

        rows = []
        for user in users:
            user_id = user.get_id()
            if not user_id.endswith('(default)'):
                user_informations = user.get_subject_informations()
                user_perimeters = user_informations.get('perimeters')
                organization_value = ''
                for perimeter in user_perimeters:
                    if perimeter.get('dimension') == filter_id:
                        organization_value = perimeter.get('valeures')
                        break
                rows.append({
                    self.username_head: user_id, 
                    self.ou_head: organization_value
                })

        # # TODO factorize following lines with CSVRightsManagement
        try:
            with open(csv_path, "w+", encoding='utf8', newline = '') as csv_file:
                writer = self.get_csv_writer(csv_file, header)
                writer.writeheader()
                for row in rows:
                    # Get the data that interest us (= filtering values and if they're activated)
                    writer.writerow(row)
                        
            print('Done ! Csv file containing all filtering values was generated at', csv_path)
        except IOError as e:
            print(e)

    def read_and_apply(self, csv_path):
        """Parse a csv file with fdu, groups and filtering values.
        Does so in order to associate the filtering values in Orion

        :param csv_path: path to the csv files
        """
        print('Scanning file', csv_path, '....')

        storage_resource = self.orion.businesses.get_stats_resource(self.storage_id)

        if storage_resource is None:
            raise Exception('Storage resource not found: {}'.format(self.storage_id))

        fdu_filter = self.orion.filters.get_with_id(storage_resource.filter_id)

        if storage_resource is None:
            raise Exception('Filter on storage resource not found: {}'.format(storage_resource.filter_id))

        # TODO Call methods using requests (call set_filter_values ?) at the end of the loop
        try:
            with open(csv_path, "r", encoding='utf8') as csv_file:
                reader = self.get_csv_reader(csv_file)
 
                next(reader)
                for row in reader:
                    username = row[self.username_i]
                    ou = row[self.ou_i]
                    if not username or not ou:
                        print('\tAssociate user with ou can\'t start because data is missing "username = {}" and "ou = {}"...'.format(username, ou))
                        continue
                    
                    user_id = username
                    if not username.endswith('(agol)'):
                        user_id = username + "(agol)"
                    user = self.orion.users.get_with_id(user_id)

                    if user is None:
                        print('\tUser {} not found '.format(username))
                        continue
                    
                    if not fdu_filter.is_label_defined(ou):
                        print('\tFilter name {} not found '.format(ou))
                        continue

                    filter_values = [ou]
                    user.set_filter_values(fdu_filter, filter_values)

            print('Done ! All possible organisation unit and user association were successfully modified !')
        except IOError as e:
            print(e)



import json
from ..RequestManager import RequestManager
from ..UrlBuilder import UrlBuilder


class Items:

    def __init__(self, type, tags, construct):
        self.type = type
        self.tags = tags
        self.url_manager = UrlBuilder()
        self.request = RequestManager()
        self.construct = construct
        self.query_supplement = "tags:" + self.tags + " AND type:" + self.type

    def _verify(self, req, true, false):
        """
        Check if the query is well executed
        :param req: result of the query
        :param true: message if the query is successful
        :param false: message if the query is not successful
        """
        if "error" in req.keys():
            if isinstance(req["error"], dict):
                if "message" in req["error"].keys():
                    raise ValueError(req["error"]["message"])
            raise ValueError(req["error"])
        if "success" in req.keys():
            if req["success"]:
                print(true)
                return
            else:
                raise ValueError(false)
        print(true)

    def search(self, q="", start=-1, num=100, all=False):
        """
        Method to search items with a query q
        :param q: query
        :param start: number of the first result, default value : -1
        :param num: number of results, default value : 100
        :param all: True to search in portal and elasticsearch, default: False to search in portal
        :return: results of the query (json), containing a list of items
        """
        result = []
        if num <= 0:
            raise ValueError("num must be > 0")
        count = 0

        #url construction
        if all is True:
            url = self.url_manager.search_all_feature_url(self.type)
        else:
            url = self.url_manager.search_feature_url(self.type)

        #request parameters
        data_q = self.query_supplement
        if len(q.strip()) != 0:
            data_q = " AND " + data_q
        data = {"start": start, "num": num, "q": q + data_q}

        #request
        while True:
            req = self.request.get(url, data).json()
            for feat in req["results"]:
                result.append(self.construct(feat))
                count += 1
                if count == num:
                    if len(result) == 0:
                        raise ValueError("No item found with this query : " + q)
                    return result
            if req["nextStart"] == -1:
                if len(result) == 0:
                    raise ValueError("No item found with this query : " + q)
                return result
            data["start"] = req["nextStart"]

    def data(self, itemId):
        """
        Method to get the data of the item with the id 'itemId'
        :param itemId: id of the item (geonote or project) you want the data
        :return: Geonote or Project, depend on the object that call this method
        """
        url = self.url_manager.data_feature_url(itemId, self.type)
        req = self.request.get(url).json()
        print("Item : "+itemId)
        self._verify(req, "Data found", "Data NOT found")
        result = json.dumps(req, indent=4, ensure_ascii=False)
        return result

    def data_export(self, itemId, export_file):
        """
        Method to get the data of the item with the id 'itemId' and export it in a json file
        It is necessary to add '.json' at the end of the path
        :param itemId: id of the item (geonote or project) you want the data
        :param export_file: path of the file where to export the data
        :return: Geonote or Project, depend on the object that call this method
        """
        if export_file[-5:] != ".json":
            raise ValueError("File must end with '.json'")
        data = self.data(itemId)
        try:
            file = open(export_file, 'w', encoding="utf-8")
            file.write(data)
            file.close()
        except:
            error = "Cannot write in file " + export_file
            print(error)
        return data


    def add_item(self, owner, title, text, snippet=None):
        """
        Add an item in aob
        :param title: title of the item
        :param text: content for the item
        :param snippet:
        :return: result of the query (a boolean for "success", the id of the item and the id of the folder)
        """
        url = self.url_manager.add_item_feature_url(self.type, owner)
        data = {"title": title, "snippet": snippet, "text": text, "type": self.type, "tags": self.tags}
        req = self.request.post(url, data).json()
        self._verify(req, "Item \"" + str(title) + "\" was added successfully.\nitemId : " + req["id"],
                     "WARNING : Item \"" + str(title) + "\" was NOT added successfully")
        return req

    def add_item_from_file(self, title, json_path, snippet=None):
        """
        Add an item in aob from a json file
        :param json_path: path of the json file
        :return: result of the query (a boolean for "success", the id of the item and the id of the folder)
        """
        with open(json_path, 'r', encoding='utf-8') as file:
            json_data = json.dumps(json.load(file), indent=4, ensure_ascii=False)
        print(json_data)
        return self.add_item(title, json_data, snippet)

    def update_item(self, itemId, title, text, snippet=None):
        """
        Update an item in aob
        :param itemId: id of the item you want to update
        :param title: new title of the item
        :param text: content for the item to be updated
        :param snippet:
        :return: result of the query (a boolean for "success" and the id of the item)
        """
        url = self.url_manager.update_item_feature_url(self.type, self._get_owner_from_id(itemId), itemId)
        data = {"title": title, "snippet": snippet, "text": text, "type": self.type, "tags": self.tags}
        req = self.request.post(url, data).json()
        self._verify(req, "Item with id: \"" + str(itemId) + "\" was updated successfully\nTitle : " + title,
                     "WARNING : Item with id: \"" + str(itemId) + "\" was NOT updated successfully")
        return req

    def update_item_from_file(self, itemId, title, json_path, snippet=None):
        """
        Update an item in aob from a json file
        :param json_path: path of the json file
        :return: result of the query (a boolean for "success", the id of the item and the id of the folder)
        """
        with open(json_path, 'r', encoding='utf-8') as file:
            json_data = json.dumps(json.load(file), indent=4, ensure_ascii=False)
        return self.update_item(itemId, title, json_data, snippet)

    def _get_owner_from_id(self, itemId):
        """
        Return the owner of the item that correspond to the item id 'itemId'
        :param itemId: id of the item you want the owner
        :return: owner of the item
        """
        return self._get_info_from_id(itemId, "owner")

    def _get_info_from_id(self, itemId, info):
        """
        Return the info wanted of the item that correspond to the itemId
        :param itemId: id of the item you want the owner
        :return: owner of the item
        """
        search = self.search(q="id:"+itemId)
        if len(search) == 0 or self.tags not in search[0].get("tags"):
            error = "There is no " + self.tags[4::] + " with this id"
            raise ValueError(error)
        result = search[0].get(info)
        return result

    def delete(self, itemId):
        """
        Delete the item with the id itemId
        :param itemId: id of the item you want to delete
        :return: result of the query (a boolean for "success" and the id of the item)
        """
        url = self.url_manager.delete_feature_url(self.type, self._get_owner_from_id(itemId), itemId)
        req = self.request.post(url).json()
        self._verify(req, "Item with id: \"" + str(itemId) + "\" was deleted successfully",
                     "WARNING : Item with id: \"" + str(itemId) + "\" was NOT deleted successfully")
        return req

    def reassign(self, itemId, targetUsername):
        """
        Change the owner of the item with the id itemId for the user "targetUsername")
        :param itemId: id of the item you want to reassign
        :param targetUsername: username of the user you want to reassign the item
        :return: result of the query (a boolean for "success" and the id of the item)
        """
        url = self.url_manager.reassign_feature_url(self.type, self._get_owner_from_id(itemId), itemId)
        data = {"targetUsername": targetUsername}
        req = self.request.post(url, data).json()
        self._verify(req, "Item with id: \"" + str(itemId) + "\" was reassigned successfully\nNew owner : " + str(targetUsername),
                     "WARNING : Item with id: \"" + str(itemId) + "\" was NOT reassigned successfully")
        return req



# https://front.arcopole.fr/orion_fed_105/orion/admin/tree/rightmanagement/groups/__children
import abc

from .Elements import Elements
from .Subject import Subject


class Subjects(Elements):
    """
    Class allowing to get access to the defined groups

    """

    def __init__(self, subject_type):
        super().__init__()
        
        self.subject_type = subject_type
        self.title_key = 'title'
        self.name_key = 'name'


    @abc.abstractmethod
    def _get_system_subject_id(self):
        """Return list of subject id"""
        raise NotImplementedError('class must define _get_system_subject_id() to use this base class')

    def _update(self):
        """Make sure that list of elements is up to date with DB state.
        Name is the key to get a particular group
        """
        subject_structs = self.request.get_in_python(self.url_manager.subject_list_url(self.subject_type))
        self._elements = {}
        for subject_struct in subject_structs:
            # Groups 'organisation' and 'tout le monde' require a specific treatment
            if subject_struct[self.name_key] in self._get_system_subject_id():
                subject_url = self.url_manager.subject_information_url(self.subject_type, subject_struct[self.name_key])
                subject_str = self.request.get_in_python(subject_url)
                profile_id = subject_str['properties']['builtinRole']
            else:
                profile_id = None
            
            subject_id = subject_struct[self.name_key]
            subject_title = subject_struct[self.title_key]

            self._elements[subject_title] = self._create_subject_instance(subject_title, subject_id, profile_id)

    @abc.abstractmethod
    def _create_subject_instance(self, subject_title, subject_id, profile_id):
        """Create instance of the specific subject"""
        raise NotImplementedError('class must define _create_subject_instance() to use this base class')

    def get_with_id(self, subject_id):
        """Look for a particular group in the list using its id

        :param group_id: id of the group to look for
        :return: the required group or None if nothing found
        """
        self._update()
        subject_id = subject_id.strip()
        for key, subject in self._elements.items():
            if subject.id == subject_id:
                return subject
        return None

    def all(self):
        """
        :return: the list of elements' values
        """
        # TODO convert to list ?
        self._update()
        return sorted(self._elements.values(),
                      key = lambda group: group.title.lower())



# https://integration.arcopole.fr/Orion/orion/admin/tree/object/BUSINESS/Cadastre

import json
from .Resource import RightLevel

from .BusinessResource import BusinessResource


# TODO : factorize with Resource !
class CadastreResource(BusinessResource):
    def __init__(self, resource_description):
        """

        :param resource_description: Description of the cadastre resource
        """
        super().__init__(resource_description)
        definition = resource_description["definition"]
        
        self._associated_filter = None
        self._associated_filter_id = definition['filterByCommunesId']
        self.associated_layer_id = definition['parcellesLayerReferenceIds']
        self.associated_service = definition['serviceReference']
        self.formated_service = self._get_formated_service(self.associated_service)
        self._rights = {"nominatif": RightLevel.NOMINATIF_ACCESS, "public": RightLevel.PUBLIC_ACCESS, "access": RightLevel.ACCESS}

    def _get_formated_service(self, url):
        if url == None or url == "":
            return url
        split_url = url.split('/')
        result = ""
        for i in range(0, len(split_url)-1):
            result += split_url[i]+'/'
        result = result[:-1]+'_'+split_url[len(split_url)-1]
        return result

    @property
    def associated_filter_id(self):
        """
        :return: Associated filter's id
        """
        # TODO get the full filter instead ! (How ?)
        return self._associated_filter_id

    @property
    def associated_filter(self):
        """
        :return: Associated filter (if init_filter_access was called)
        """
        if self._associated_filter is None:
            print(
                "To get associated_filter, you must call init_filter_access(associated_filter)",
                "method first, using associated_filter_id")
            return None
        return self._associated_filter

    def init_filter_access(self, associated_filter):
        """Initialize filter access

        :param associated_filter: FDU corresponding to self.associated_filter_id
        """
        # TODO : method "get_filter_access(orion)" ?
        if associated_filter.get_id() == self.associated_filter_id:
            self._associated_filter = associated_filter
        else:
            print("ERROR : you called init_filter_access not with the exact associated filter")

    # ----- Update level of right -----

    def update_right(self, group, new_right):
        """If possible, update the group's right level on this resource

        :param group: group concerned
        :param new_right: new right level (RightLevel type)
        """
        if not self._can_modify(group, new_right):
            return
        self._apply_new_right(group, new_right)

    def update_right_from_string(self, group, right_string):
        """Call the method update_right() if the right is good

        :param group: group concerned
        :param right_string: right formatted to string
        :return: True if the right is good, False if it's not good
        """
        if not self._is_right_good(right_string):
            print("[WARNING] There's an error with the right, it will not be updated")
            return False
        self.update_right(group, self._rights[right_string])
        return True

    def _is_right_good(self, right_string):
        """
        Check if the right string is good

        :param right_string:
        :return: False if the right string is not good, else return True
        """
        if right_string not in self._rights.keys():
            return False
        return True

    # ----- Activate a filter -----

    def activate_filter(self, group):
        """Activates the associated filter for a given group

        :param group: group to activate filter for
        """
        current_rights = self._get_right_structure(group)
        if self.associated_filter is None:
            print("You must call CadastreResource:init_filter_access before !")
            return
        if not self._can_activate_filter(current_rights, self.associated_filter, group):
            return
        self._activate_filter(current_rights, self.associated_filter, group)
        if not group.has_defined_filter_values(self.associated_filter):
            print('WARNING : you should define filtering values for this group and this filter')

    def _can_activate_filter(self, group_rights, filt, group=None):
        """Verify if can apply a given filter to this resource

        :param group_rights: Contains current rights on this resource
        :param group: the group for which apply the filter
        :return: True if all tests are passed. False otherwise
        """
        # TODO : also call _can_modify ?
        if not group_rights['rights']:
            print('The resource has no PUBLIC_ACCESS nor NOMINATIF_RIGHT right = can\'t activate filter.'
                  '1 FIX idea=Change right resource.update_rights(group, RightLevel.[PUBLIC_ACCESS|NOMINATIF_ACCESS])')
            return False
        if filt.get_id() in group_rights['rights'][0]['filteredDimensions']:
            print('Filter already applied, nothing to change')
            return False
        return True

    def _activate_filter(self, current_rights, filt, group):
        """Activate a filter for a group on this resource

        :param group: group for which apply filter
        """
        current_rights['rights'][0]['filteredDimensions'].append(filt.get_id())
        config_url = self._url_builder.cadastre_configuration_url(group.get_id())
        self._request_mgr.post(config_url, data = {'value': json.dumps(current_rights)})
        # print(req, req.text)
        print('Filter successfully applied on cadastre resource ',
              'for group', group.get_name())

    # ----- Deactivate a filter -----

    def deactivate_filter(self, group):
        """Deactivates the associated filter for a given group

        :param group: group to deactivate filter for
        """
        if self.associated_filter is None:
            return
        self._deactivate_filter(self.associated_filter, group)

    def _deactivate_filter(self, filt, group):
        """Remove a filter for a group on this resource

        :param filt: Filter to deactivate
        :param group: group for which disable filter"""
        current_rights = self._get_right_structure(group)
        if not self._can_deactivate_filter(current_rights, filt, group):
            return
        current_rights['rights'][0]['filteredDimensions'].remove(filt.id)
        config_url = self._url_builder.cadastre_configuration_url(group.get_id())
        self._request_mgr.post(config_url, data = {'value': json.dumps(current_rights)})
        # print(req, req.text)  # log
        print('Filter', filt.name, 'successfully deactivated on cadastre resource for',
              group.get_name())

    def _can_deactivate_filter(self, current_rights, filt, group):
        """Verify if can deactivate a given filter to this resource

        :param current_rights: Contains current rights on this resource
        :param filt: filter to deactivate
        :param group: the group for which deactivate the filter
        :return: True if all tests are passed. False otherwise
        """
        if not self._can_modify(group = group, right = ""):
            return False
        if not current_rights['rights']:
            print('The resource has no PUBLIC nor NOMINATIF right = no filter to deactivate')
            return False
        if current_rights['rights'][0]['action'] == RightLevel.NOMINATIF_ACCESS.value:
            print('The resource has a NOMINATIF right. Filter deactivation is not possible')
            return False
        # NB : as we use associated_filter, the following should never happen. (keep it ?)
        if filt.id not in current_rights['rights'][0]['filteredDimensions']:
            print('Filter to deactivate not defined for this group on this resource')
            return False
        return True

    def _can_modify(self, group, right):
        """

        :param group:
        :param right:
        :return: True if can modify and apply given right. False otherwise.
        """
        # TODO : check if right is NOMINATIF_access, public_access or access ?
        # TODO : create new method can_change_rights (cf Resource)
        if not self.is_shared_with(group):  # TODO doesn't work yet
            print('Cadastre resource must be shared with the group {gr} '
                  'to change anything'.format(gr = group.get_name()))
            return False
        if right == RightLevel.NOMINATIF_ACCESS:
            if self.associated_filter is None:
                print("You must call CadastreResource:init_filter_access before !")
                return False
            if not group.has_defined_filter_values(self.associated_filter):
                print("{gr} must have at least one filter value defined to "
                      "apply NOMINATIF_ACCESS right".format(gr = group.get_name()))
                return False
        return True

    def _apply_new_right(self, group, new_right):
        if new_right == RightLevel.NOMINATIF_ACCESS:
            self.activate_filter(group)
        current_rights = self._prepare_rights_structure(self._get_right_structure(group),
                                                        new_right)
        self._request_mgr.post(
            self._url_builder.cadastre_configuration_url(group.profile_id),
            data = {'value': json.dumps(current_rights)})
        print('Switched to', new_right.name, 'for', group.get_name(),
              'on cadastre resource !')

    def _prepare_rights_structure(self, current_rights, right_level):
        """Prepare structure that will be used for setting rights

        :param current_rights: the existing structure to modify. Allows to save applied filters
        :param right_level: new right level
        :return: the structure built.
        """
        if current_rights['rights']:  # if rights already has fields
            if right_level == RightLevel.PUBLIC_ACCESS or right_level == RightLevel.NOMINATIF_ACCESS:
                current_rights['rights'][0]['action'] = right_level.value
            elif right_level == RightLevel.ACCESS:
                current_rights['rights'] = []
        else:  # requires to build structure
            if right_level == RightLevel.PUBLIC_ACCESS:
                rights_dic = {'filteredDimensions': [],
                              'action': right_level.value}
                current_rights['rights'].append(rights_dic)
            elif right_level == RightLevel.NOMINATIF_ACCESS:
                rights_dic = {'action': right_level.value}
                current_rights['rights'].append(rights_dic)
        return current_rights

    def _get_right_structure(self, group):
        """ Gives the cadastre resource current rights for a group

        :param group: group for which we want the rights
        :return: a structure (dictionary) representing resource current rights
        """
        resource_rights_url = self._url_builder.cadastre_rights_url(profile_id = group.profile_id)

        req = self._request_mgr.get_in_python(resource_rights_url)
        # print('current :', req)
        return req

    def print_rights(self, group):
        """Prints current rights for a given group

        :param group: The group for which we want the rights
        """
        rights = self._get_right_structure(group)
        rights_str = ['Here are the rights for ', group.get_name(),
                      ' on the cadastre resource : \n']
        if rights['rights']:
            rights_str.append("\t-The right is ")
            rights_str.append(rights['rights'][0]['action'].upper() + '\n')
            if rights['rights'][0]['filteredDimensions']:
                rights_str.append("\t-Filter(s) applied :\n")
                for filt in rights['rights'][0]['filteredDimensions']:
                    rights_str.append("\t\t\t*" + filt + "\n")
        else:
            rights_str.append('\t-The right is ACCESS\n')
        print("".join(rights_str))

    def is_shared_with(self, group):  # TODO implement
        """Check if the resource is shared with a given group

        :param group: the group to check for
        :return: True if resource shared with group, false otherwise.

        .. warning:: For now, doesn't work"""
        return True

    def __str__(self):
        """Provides a string representation of a cadastre resource"""
        resource_str = 'Resource id : {}, ' \
                       'Service associé : {}, ' \
                       'Service associé formaté : {}, ' \
                       'Filtre associé : {}'.format(self.id,
                                                    self.associated_service,
                                                    self.formated_service,
                                                    self._associated_filter_id)
        return resource_str



# *** Obtenir les droits résolus ***
# https://{orionHeadAdmin}/rightmanagement/groups/{group_id}/authorizedResources/SERVICES/{resourcePath}

# *** Obtenir les droits configurés ***
# https://{orionHeadAdmin}/rightmanagement/profiles/{profileId}/aobrights/SERVICES/{resourcePath}


import abc
import json
from enum import Enum, unique

from .. import cfg_global
from ..RequestManager import RequestManager
from ..Filter import _FDUFilter, _SQLFilter
from ..UrlBuilder import UrlBuilder


@unique  # Ensures that enum's values are unique
class RightLevel(Enum):
    """An enumeration.
    Represents the different levels of rights which are possible to apply
    The values being the ones used in the request."""
    # TODO handle RightLevel.NO_ACCESS
    NO_ACCESS = "nothing"
    ACCESS = "access"
    READ = "read"
    WRITE = "write"
    PUBLIC_ACCESS = "public_access"
    NOMINATIF_ACCESS = "nominatif_access"


class Resource(metaclass = abc.ABCMeta):
    """
    Common abstract class for resources (service, layer, ...)
    """

    # TODO factorize prints and .format()

    def __init__(self, access_url, capabilities="", parent_access="", is_managed=False):
        self.access_url = access_url
        self.capa = capabilities
        self.parent_access = parent_access
        self._url_builder = UrlBuilder()
        self._request_mgr = RequestManager()
        self._is_managed = is_managed
        self.name = "TMP"
        self.type = ""
        self.groups = []

    # ----- Activate a filter -----

    def activate_fdu_filter(self, group, filt):
        # TODO method force_fdu_filter(group, filter) ?
        """Add a FDU filter for a group on this resource

        :param filt: The filter (FDU) to apply
        :param group: group for which apply filter
        """
        current_rights = self._get_right_structure(group)
        if not self._can_activate_FDU(current_rights, filt, group):
            return
        self._activate_filter(current_rights, filt, group)
        if not group.has_defined_filter_values(filt):
            print('WARNING : you should define filtering values for this group and this filter')

    def activate_sql_filter(self, group, filt):
        # TODO method force_sql_filter(group, filter) ?
        """Add a SQL filter for a group on this resource

        :param filt: The filter (FDU) to apply
        :param group: group for which apply filter
        """
        current_rights = self._get_right_structure(group)
        if not self._can_activate_SQL(current_rights, filt, group):
            return
        self._activate_filter(current_rights, filt, group)

    def _activate_filter(self, current_rights, filt, group):
        """Activate a filter for a group on this resource

        :param filt: The filter to apply
        :param group: group for which apply filter
        """
        current_rights['rights'][0]['filteredDimensions'].append(filt.id)
        config_url = self._url_builder.resource_configuration_url(group.get_id(), self.access_url)
        self._request_mgr.post(config_url, data = {'value': json.dumps(current_rights)})
        # print(req, req.text)
        print('Filter', filt.name, 'successfully applied on resource', self.access_url,
              'for group', group.get_name())

    # ----- Deactivate a filter -----

    def deactivate_fdu_filter(self, group, filt):
        """Remove a (FDU) filter for a group on this resource

        :param filt: Filter to deactivate
        :param group: group for which disable filter"""
        self._deactivate_filter(filt, group)

    def deactivate_sql_filter(self, group, filt):
        """Remove a (SQL) filter for a group on this resource

        :param filt: Filter to deactivate
        :param group: group for which disable filter"""
        self._deactivate_filter(filt, group)

    def _deactivate_filter(self, filt, group):
        """Remove a filter for a group on this resource

        :param filt: Filter to deactivate
        :param group: group for which disable filter"""
        current_rights = self._get_right_structure(group)
        if not self._can_deactivate_filter(current_rights, filt, group):
            return
        current_rights['rights'][0]['filteredDimensions'].remove(filt.id)
        config_url = self._url_builder.resource_configuration_url(group.get_id(), self.access_url)
        self._request_mgr.post(config_url, data = {'value': json.dumps(current_rights)})
        # print(req, req.text)  # log
        print('Filter', filt.name, 'successfully deactivated on resource', self.access_url,
              'for', group.get_name())

    # ----- Access to current/parent rights -----

    def _get_right_structure(self, group):
        """ Gives the resource current rights for a group

        :param group: group for which we want the rights
        :return: a structure (dictionary) representing resource current rights
        """
        resource_rights_url = self._url_builder.resource_rights_url(profile_id = group.profile_id,
                                                                    resource_path = self.access_url)

        req = self._request_mgr.get_in_python(resource_rights_url)
        # print('current :', req)
        return req

    def get_resolved_right(self, group):
        """Returns the resolved right

        :param group: group for which we want the rights
        :return: The right level's value
        """
        resolved_url = self._url_builder.resource_resolved_permissions_url(group.get_id(),
                                                                           self.access_url)
        resolved_rights = self._request_mgr.get_in_python(resolved_url)
        resolved_perm_key = 'resolvedPermissions'
        if RightLevel.WRITE.value in resolved_rights[resolved_perm_key] \
                and resolved_rights[resolved_perm_key][RightLevel.WRITE.value]['permission']:
            return RightLevel.WRITE
        if RightLevel.READ.value in resolved_rights[resolved_perm_key] \
                and resolved_rights[resolved_perm_key][RightLevel.READ.value]['permission']:
            return RightLevel.READ
        return RightLevel.ACCESS

    def print_rights(self, group):
        """Prints current rights for a group

        :param group: The group for which we want the rights
        """
        rights = self._get_right_structure(group)
        rights_str = ['Here are the rights for ', group.get_name(), ' on ',
                      self.access_url, ' : \n']
        if rights['isInherited']:
            rights_str.append("\t-Right inherited\n")
            rights_str.append("\t-The right is ")
            rights_str.append(self.get_resolved_right(group).value.upper() + '\n')
        else:
            rights_str.append("\t-Right not inherited\n")
            if rights['rights']:
                rights_str.append("\t-The right is ")
                rights_str.append(rights['rights'][0]['action'].upper() + '\n')
                if rights['rights'][0]['filteredDimensions']:
                    rights_str.append("\t-Filter(s) applied :\n")
                    for filt in rights['rights'][0]['filteredDimensions']:
                        rights_str.append("\t\t\t*" + filt + "\n")
            else:
                rights_str.append('\t-The right is ACCESS\n')
        print("".join(rights_str))

    def _get_parent_rights(self, group):
        """ Get a resource parent's current rights for a group

        :param group: id of the group for wich we want information
        :return: the resource parent's rights for this group
        """
        req = self._request_mgr.get_in_python(
            self._url_builder.resource_rights_url(group.get_id(), self.parent_access))
        print('parent :', req)
        return req

    # ----- Update level of right -----

    def clear_all_rights(self):
        """
        If possible, clears given rights and filters on a resource.
        However, it will not remove sharing accesses.
        """
        if not self.is_managed():
            return
        self._request_mgr.post(
            self._url_builder.resource_clear_all_url(self.access_url)
        )
        print('Cleared all rights on resource ', self.access_url)

    def force_right(self, group, new_right):
        """Method forcing a right on a resource

        :param group: group concerned
        :param new_right: new right level (RightLevel type)
        """
        # Enables right handling if not already done.
        if cfg_global.is_federated and not self.is_managed() and self.get_type() == "SERVICE":
            self.enable()

        if not self._can_modify(group, check_inheritance = False) or \
                not self._can_change_rights(group, new_right):
            # print('Impossible to change right but will at least try to disable inheritance')
            self.disable_inheritance(group)
            return

        # Applies the new configuration
        self._apply_new_right(group, new_right, disable_inheritance = True)

    def update_right(self, group, new_right):
        """If possible, update the group's right level on this resource

        :param group: group concerned
        :param new_right: new right level (RightLevel type)
        """
        # Check if can modify anything about the rights
        if not self._can_modify(group) or \
                not self._can_change_rights(group, new_right):
            return
        # Applies the new configuration
        self._apply_new_right(group, new_right)

    def _apply_new_right(self, group, new_right, disable_inheritance=False):
        """Prepares right structure and apply right modification on the resource

        :param group: group concerned
        :param new_right: new right level (RightLevel type)
        :param disable_inheritance: If it is required to disable inheritance or not
        :return:
        """
        current_rights = self._prepare_rights_structure(self._get_right_structure(group),
                                                        new_right,
                                                        disable_inheritance)
        self._request_mgr.post(
            self._url_builder.resource_configuration_url(group.profile_id, self.access_url),
            data = {'value': json.dumps(current_rights)})
        # print(answer, answer.text)  # log
        print('Switched to', new_right.name, 'for', group.get_name(),
              'on resource', self.access_url)

    def _prepare_rights_structure(self, current_rights, right_level, disable_inheritance=False):
        """Prepare structure that will be used for setting rights

        :param current_rights: the existing structure to modify. Allows to save applied filters
        :param right_level: new right level
        :return: the structure built.
        """
        if current_rights['rights']:  # if rights already has fields
            if right_level == RightLevel.READ or right_level == RightLevel.WRITE:
                current_rights['rights'][0]['action'] = right_level.value
            elif right_level == RightLevel.ACCESS:
                current_rights['rights'] = []
        else:  # requires to build structure
            if right_level == RightLevel.READ or right_level == RightLevel.WRITE:
                rights_dic = {'filteredDimensions': [],
                              'action': right_level.value}
                current_rights['rights'].append(rights_dic)
        if disable_inheritance:
            current_rights['isInherited'] = False
        return current_rights

    # ----- Tests validity of modification -----

    def is_shared_with(self, group):  # TODO implement this for every resource type
        """Check if the resource is shared with a given group

        :param group: the group to check for
        :return: True if resource shared with group, false otherwise.

        .. warning:: For now, only work for services"""
        # si id groupe dans groups, true else false
        return True

    def _has_parent(self):
        """Check if the resource has a parent

        :return: if no parent (a service), returns an empty string(=False).
        Otherwise, returns parent access (non-emptystring/True)
        """
        return self.parent_access

    def is_managed(self):
        """Check if resource's rights are managed by aOB

        :return: True if resource managed by aOB, False otherwise."""
        if not cfg_global.is_federated:
            return True  # avoid problems during right configuration
        return self._is_managed

    def has_inherited_right(self, group):
        """Checks if right is inherited for group

        :param group: group for which look for
        :return:
        """
        current_rights = self._get_right_structure(group)
        return current_rights['isInherited']

    def _can_modify(self, group, check_inheritance=True):
        """Verify if can change anything about rights on this resource

        :param group: Group for which modify anything about rights
        """
        if not self.is_managed():
            print('You must enable {rs} to handle rights first : '
                  'Use Service::enable()'.format(rs = self.get_name()))
            return False
        if not self.is_shared_with(group):  # TODO doesn't work yet
            print('{rs} must be shared with the group {gr} '
                  'to change anything'.format(rs = self.get_name(), gr = group.get_name()))
            return False
        if check_inheritance and self.has_inherited_right(group):
            print("You must disable inheritance on {rs} for group {gr} to change anything : "
                  "Use the disable_inheritance(group) method".format(rs = self.get_name(),
                                                                     gr = group.get_name()))
            return False
        return True

    def _can_change_rights(self, group, right_to_apply):
        """Verify if can change the right level for this resource

        :param group: id of the group concerned by this modification
        :param right_to_apply: New right level to apply
        """
        if right_to_apply == RightLevel.WRITE \
                and self.capa and 'Editing' not in self.capa:
            print('Impossible to add editing write '
                  'for group {gr} on {rs}'.format(gr = group.get_name(),
                                                  rs = self.get_name()))
            return False

        current_right = self.get_resolved_right(group)
        if current_right == right_to_apply:
            # print('Right already defined as', right_to_apply.value.upper(),
            #       'for group {gr} on {rs}, nothing to change'.format(rs = self.get_name(),
            #                                                          gr = group.get_name()))
            return False
        return True

    # ----- Tests validity to (de)activate filter -----

    def _can_activate_SQL(self, group_rights, filt, group):
        """Verify if can apply a SQL filter to this resource

        :param group_rights: Contains current rights on this resource
        :param filt: filter to apply
        :param group: the group for which apply the filter
        :return: True if all tests are passed. False otherwise
        """
        if not isinstance(filt, _SQLFilter):
            print('Filter must be of SQL type')
            return False
        if not self._can_activate_filter(group_rights, filt, group):
            return False
        return True

    def _can_activate_FDU(self, group_rights, filt, group):
        """Verify if can apply a FDU filter to this resource

        :param group_rights: Contains current rights on this resource
        :param filt: filter to apply
        :param group: the group for which apply the filter
        :return: True if all tests are passed. False otherwise
        """
        if not isinstance(filt, _FDUFilter):
            print('Filter must be of FDU type')
            return False
        if not self._can_activate_filter(group_rights, filt, group):
            return False
        return True

    def _can_activate_filter(self, group_rights, filt, group):
        """Verify if can apply a given filter to this resource

        :param group_rights: Contains current rights on this resource
        :param filt: filter to apply
        :param group: the group for which apply the filter
        :return: True if all tests are passed. False otherwise
        """
        if not self._can_modify(group = group):
            return False
        if not group_rights['rights']:
            print('The resource has no READ nor WRITE right = no filter to activate.'
                  '1 FIX idea= Change right resource.update_rights(group, RightLevel.[READ|WRITE])')
            return False
        if filt.id in group_rights['rights'][0]['filteredDimensions']:
            print('Filter already applied, nothing to change')
            return False
        return True

    def _can_deactivate_filter(self, current_rights, filt, group):
        """Verify if can deactivate a given filter to this resource

        :param current_rights: Contains current rights on this resource
        :param filt: filter to apply
        :param group: the group for which apply the filter
        :return: True if all tests are passed. False otherwise
        """
        if not self._can_modify(group = group):
            return False
        if not current_rights['rights']:
            print('The resource has no READ nor WRITE right = no filter to activate')
            return False
        if filt.id not in current_rights['rights'][0]['filteredDimensions']:
            print('Filter to deactivate not defined for this group on this resource')
            return False
        return True

    # ----- Change inheritance -----

    def enable_inheritance(self, group):
        """Enables inheritance

        :param group: the group for which change inheritance
        """
        self._change_inheritance(group, True)

    def disable_inheritance(self, group):
        """Disable inheritance

        :param group: the group for which change inheritance
        """
        self._change_inheritance(group, False)

    def _change_inheritance(self, group, new_inheritance):
        """Disable or enable inheritance of rights for the group on this resource

        :param group: the group for which change inheritance
        :param new_inheritance: a boolean. True if inherit, false otherwise
        """
        if not self._can_modify(group, check_inheritance = False):
            return False
        group_rights = self._get_right_structure(group)
        if group_rights['isInherited'] == new_inheritance:
            # print('Nothing to change in inheritance')
            return False

        group_rights['isInherited'] = new_inheritance

        if not group_rights['isInherited']:  # if disable inheritance, need to prepare structure...
            right = self.get_resolved_right(group)
            self._prepare_rights(group_rights, right)

        # Does the action of changing inheritance
        self._request_mgr.post(
            self._url_builder.resource_configuration_url(group.profile_id, self.access_url),
            data = {'value': json.dumps(group_rights)})

        inherit_str = 'enabled' if new_inheritance else 'disabled'
        print('Inheritance {inh} for {gr} on {rs}'.format(inh = inherit_str,
                                                          gr = group.get_name(),
                                                          rs = self.get_name()))

    # ----- Access methods -----
    def get_access_url(self):
        return self.access_url

    def get_name(self):
        return self.name

    @staticmethod
    def get_type():
        # TODO find another way to do this
        pass

    @staticmethod
    def _prepare_rights(group_rights, action):
        """Prepares the right structure when disable_inheritance

        :param group_rights: structure to modify
        :param action: action to apply (RightLevel type)
        """
        if action == RightLevel.ACCESS:
            group_rights['rights'] = []
        elif group_rights['rights']:
            # This allows to save filters if set
            group_rights['rights'][0]['action'] = action.value
        else:  # an empty list and needs to create structure.
            group_rights['rights'] = [{'action': action.value}]



# https://integration.arcopole.fr/Orion/orion/admin/tree/object/BUSINESS/Cadastre

import json
from .BusinessResource import BusinessResource


# TODO : factorize with Resource !
class StorageResource(BusinessResource):
    def __init__(self, resource_description):
        """

        :param resource_id: Param _id returned by the REST API
        :param definition: Definition of the cadastre resource
        """
        super().__init__(resource_description)

    @property
    def filter_id(self):
        """
        :return: Associated filter's id
        """
        # TODO get the full filter instead ! (How ?)
        return self._description.get('dimensionId')

    # ----- Activate a filter -----

    def update_filter(self, filter_id):
        """Activates the associated filter for a given group

        :param group: group to activate filter for
        """
        if self.filter_id == filter_id:
            print("The filter is already used for this stats storage")
        else:
            config_url = self._url_builder.stats_configuration_url(self.id)
            
            self._description['dimensionId'] = filter_id
            self._request_mgr.post(config_url, data = {'value': json.dumps(self._description)})
            
            print('Filter', filter_id, 'successfully updated in Stats resource')


    def __str__(self):
        """Provides a string representation of a cadastre resource"""
        resource_str = 'Resource id : {}, ' \
                       'Filtre associé : {}'.format(self.id,
                                                    self.filter_id)
        return resource_str



import abc

from .RequestManager import RequestManager
from .UrlBuilder import UrlBuilder


class Elements(metaclass = abc.ABCMeta):
    # TODO Add a "search" method ?
    """Interface for classes containing a list of elements.

    Attributes:
        _elements : Dictionary containing elements
    """

    def __init__(self):
        self._elements = {}
        self.url_manager = UrlBuilder()
        self.request = RequestManager()

    @abc.abstractmethod
    def _update(self):
        """Updates elements for it to be consistent with database"""
        raise NotImplementedError('class must define _update() to use this base class')

    def _update_one(self, element_id):
        """Updates one element for it to be consistent with database"""
        return None

    def all(self):
        """
        :return: the list of elements' values sorted case-insensitive by their name
        """
        self._update()
        return sorted(self._elements.values(), key = lambda elem: elem.name.lower())

    def get(self, element_name):
        """Look for a particular element in elements.

        :param element_name: identification of element
        :return: required element or None if nothing found
        """
        # If possible and useful, call update_one instead !
        self._update()
        element_name = element_name.strip()
        if element_name in self._elements:
            return self._elements[element_name]
        print("[WARNING] element", element_name, "doesn't exist, None is returned")
        return None



# https://front.arcopole.fr/Orion/orion/admin/tree/rightmanagement/profiles/authenticated/aobrights/SERVICES/SampleWorldCities_MapServer/__configure
# {"isInherited":false,"rights":[{"filteredDimensions":["32d0b429-452a-40bd-9865-5a6c6798007a"],"action":"read"}]}

from .CadastreResource import CadastreResource
from .StatsResource import StatsResource
from .StorageResource import StorageResource
from .. import cfg_global
from ..Elements import Elements


class Businesses(Elements):
    """
    Class allowing to get access to the list of business resource
    """

    def __init__(self):
        """Initialize and build our list of resources
        """
        super().__init__()

    # ----- All business resources in aOB -----
    def _update(self):
        """ Update the list of resources to make sure it's consistent with our database.
        """
        self._resources = {}
        self._build_all()

    def _build_all(self):
        """Method to build a list with all resources
        """
        node_info_url = self.url_manager.businesses_children_url()
        list_nodes = self.request.get_in_python(node_info_url)
        for node in list_nodes:
            print(node['nodeType'])
            if node['nodeType'] == 'ResourceNode':
                resource_name = node['name']
                self._add_resource_to_list(resource_name)

    def all(self):
        """
        :return: the list of elements' values
        """
        # TODO convert to list ?
        self._update()
        return self._resources

    # ----- Geting 1 resource handled by aOB -----
    def get(self, element_name):
        """Look if a particular element exist.

        :param element_name: identification of element
        :return: required element or None if nothing found
        """
        element_name = element_name.strip()
        element = self._update_one(element_name)
        if element is None:
            print("[WARNING] element", element_name, "doesn't exist, None is returned")
            return None
        return element

    def _update_one(self, element_id):
        """Update one resource required

        :param element_id: REST access url of the resource
        :return: The resource if found. None otherwise"""
        return self._create_resource(element_id)


     # ----- Getting the stats resource ----
    def get_stats_resource(self, storage_id="hosted"):
        """Get and return the stats resource if defined

        :param storage_id: id of storage define in stats resource configuration
        :return: the StatsResource or StorageResource or None if not defined.
        """
        stats_req = self._try_to_get_stats_resource(storage_id)

        if stats_req is None:
            return None
        
        if storage_id:
            return StorageResource(stats_req)
        return StatsResource(stats_req)

    def _try_to_get_stats_resource(self, storage_id=""):
        """Try to get the stats resource and handle exception if there is one."""
        stats_req = None
        try:
            stats_req = self.request.get_in_python(self.url_manager.stats_resource_url(storage_id))
        except Exception:
            stats_req = None

        return stats_req

    # ----- Getting the cadastre resource ----
    def get_cadastre_resource(self):
        """Get and return the cadastre resource if defined

        :return: the CadastreResource or None if not defined.
        """
        cadastre_req = self._try_to_get_cadastre_resource()

        if cadastre_req is None or not cadastre_req['definition']['serviceReference']:
            return None
        return CadastreResource(cadastre_req)

    def _try_to_get_cadastre_resource(self):
        """Try to get the cadastre resource and handle exception if there is one."""
        cadastre_req = None
        try:
            cadastre_req = self.request.get_in_python(self.url_manager.cadastre_definition_url())
        except Exception:
            cadastre_req = None

        return cadastre_req

    # ----- Methods for creation of a resource instance -----
    def _add_resource_to_list(self, resource_name):
        """Creates an instance of a resource and add it to the appropiated list

        :param resource_name:
        :return:
        """
        resource_instance = self._create_resource(resource_name)
        if resource_instance is None:
            return
        self._resources[resource_name] = resource_instance

    def _create_resource(self, resource_name):
        """Creates an instance of a resource if it is of a good type.

        :param resource_name:
        :return: The resource created or None if there was an error
        """
        resource_req = self._try_to_get_resource(resource_name)
        if resource_req is None or "error" in resource_req:
            print(resource_req)
            return None
        
        if resource_req['module'] == 'cadastre':
            return CadastreResource(resource_req)
        
        if resource_req['module'] == 'stats':
            return StatsResource(resource_req)

    def _try_to_get_resource(self, resource_name):
        """(FIX) Try to get a service and handle exception if there is one."""
        service_req = None
        try:
            service_req = self.request.get_in_python(self.url_manager.business_definition_url(resource_name))
        except Exception:
            service_req = None

        return service_req


# **** Get the list of all filters ****
#  https://front.arcopole.fr/Orion/orion/admin/tree/config/dimensions
# GET
# f & token

#  **** Creation/Edition/Duplication/Deletion of a new filter ****
#  https://front.arcopole.fr/Orion/orion/admin/tree/config/dimensions/__configure
# GET (create)
# f & token & value = list of all filters plus the one added
# For creation, the one added doesn't have the id & dbId fields set & partitions is empty

# **** Configuration des valeurs de filtrage ****
# https://front.arcopole.fr/Orion/orion/admin/tree/rightmanagement/groups/$GROUP_ID$/__configure
# GET
# f & token &
# value = {"name":$GROUP_ID$,
#           "domaine":$GROUP_DOMAIN$,
#           "perimeters":
#                       [{"dimension":$FILTER_ID$,
#                         "valeures":$ONE_SELECTED_FILTERING_VAL$
#                        }*
#                       ]
#         }
from .Exceptions import BuildFilterError


class FilterFactory:
    # TODO figure out what to do if dic AND parameters given
    # TODO is class really useful ?...
    # TODO Abstract factory ?
    """The Factory for our filter.

    Create a Filter of SQL or FDU type based on the arguments given.
    e.g. if the whereClause is given, then this might be a SQL filter
    """

    @staticmethod
    def try_to_build_sql(name, where_clause, filter_dic):
        """Internal function. Check if parameters to build a sql filter are good.
        If that's the case, build a SQL filter

        :return: the Filter built or None if nothing was built
        :raises BuildFilterError: if name or whereClause is empty
        """
        if filter_dic is not None \
                and "whereClause" in filter_dic and filter_dic['whereClause']:
            return _SQLFilter(filter_dic['name'].strip(), filter_dic['whereClause'].strip(),
                              filter_dic)
        if name is not None and where_clause is not None:
            if name.strip() and where_clause.strip():  # name and whereClause aren't empty
                return _SQLFilter(name.strip(), where_clause.strip())
            else:
                raise BuildFilterError(
                    "[ERROR SQL Filter] name and whereClause must not be empty")
        return None

    @staticmethod
    def try_to_build_fdu(name, attributes, properties, filtering_values, filter_dic):
        """Internal function. Check if parameters to build a FDU filter are good.
        If that's the case, build a FDU filter

        :return: the Filter built or None if nothing was built
        :raises: BuildFilterError if name or attributeName or partition is empty
        """
        if filter_dic is not None and "attributeName" in filter_dic:
            return _FDUFilter(filter_dic['name'], filter_dic['attributeName'],
                              filter_dic['properties'], filter_dic['partitions'],
                              filter_dic)
        if name is not None and attributes is not None and filtering_values is not None:
            if name.strip() and attributes and filtering_values:
                return _FDUFilter(name.strip(), attributes, properties, filtering_values)
            else:
                raise BuildFilterError(
                    "[ERROR FDU] name, attributeName and filtering_values must not be empty")
        return None

    def build_filter(name=None, attributes=None, properties=None, filtering_values=None,
                     where_clause=None, filter_dic=None, filter_type=None):
        # TODO Improve this method !
        """
        :param where_clause:
        :param filtering_values:
        :param properties:
        :param attributes:
        :param name:
        :param filter_dic: Alternative construction. Dictionary containing all parameters
        :param filter_type: Type of filter required (sql or fdu)
        :return: the Filter built or None if there was a problem during construction
        """
        try:
            if filter_type is not None:
                if filter_type != 'sql' and filter_type != 'fdu':
                    raise BuildFilterError("[ERROR Filter] invalid type_filt.", filter_type,
                                           "must be 'fdu' or 'sql'")
                if filter_type == 'sql':
                    return FilterFactory.try_to_build_sql(name, where_clause, filter_dic)
                if filter_type == 'fdu':
                    return FilterFactory.try_to_build_fdu(name, attributes, properties,
                                                          filtering_values, filter_dic)
            else:
                new_filter = FilterFactory.try_to_build_sql(name, where_clause, filter_dic)
                if new_filter is None:
                    new_filter = FilterFactory.try_to_build_fdu(name, attributes, properties,
                                                                filtering_values, filter_dic)
                    if new_filter is None:
                        raise BuildFilterError(
                            "[ERROR FILTER] Problem in arguments, no filter built")
        except BuildFilterError as e:
            print(e.args)
            new_filter = None
        return new_filter

    build_filter = staticmethod(build_filter)


class Filter:
    # TODO private class ?
    def __init__(self, name=None, filter_dic=None, filter_id=None):
        """Create a generic filter

        :param name: Filter name
        :param filter_id: Filter id
        :param filter_dic: A dictionary that containing filter's parameters
        """
        if filter_dic is not None:
            self.name = filter_dic['name']
            self.id = filter_dic['id']
            if "dbID" in filter_dic:
                self.db_id = filter_dic['dbId']
        else:
            self.name = name
            self.id = filter_id

    def get_id(self):
        """ Returns filter's id """
        if self.id is not None:
            return self.id
        else:
            print('No id set for this filter')
            return None

    def get_name(self):
        return self.name

    def __str__(self):
        """Allows to have a string representation of a filter to print.

        :return: the string representation of a filter.
        """
        return 'Filter {}; id {}'.format(self.name, self.id)


class _SQLFilter(Filter):
    """SQLFilter. A specific type of filter with a sql where request.
    SQL Filter can be applied only to a layer or a table (not a service)
    """

    def __init__(self, name="", whereClause="", filter_dic=None, filter_id=None):
        super(_SQLFilter, self).__init__(name = name,
                                         filter_dic = filter_dic,
                                         filter_id = filter_id)
        if filter_dic is not None:
            self.whereClause = filter_dic['whereClause'].strip()
        else:
            self.whereClause = whereClause.strip()

    def __str__(self):
        filter_str = ['SQL Filter "{name}"; whereClause : {wc}'.format(name = self.name,
                                                                       wc = self.whereClause)]
        if self.id is not None:
            filter_str.append("; id {}".format(self.id))
        return "".join(filter_str)


class _FDUFilter(Filter):
    """
    A filter depending on the user. Applied on some specific values of attributes.
    Can be applied only if fields of the filter are in the layer/table selected
    Can be applied on a layer only if applied on the corresponding service/table
     {
      "name":$FILTER_NAME$,
      "id": $ID$,
      "dbId":$DB_ID$,
      "attributeName":
       [
          [$fieldName$]+
       ],
      "properties":null,
      "partitions":
       [
          [{
           "name":$name$,
           "properties":
            {
             [$label$: $associatedValue$]+
            }
          }]+
       ],
      "whereClause":null
     },
    """

    def __init__(self, name, attributeName, properties=None,
                 partitions=None, filter_dic=None, filter_id=None):
        """
        :param attributeName: Field(s) concerned by the filter
        :param properties: ??
        :param partitions: valeurs de filtrage. List of value(s) selected for the field
        """
        super(_FDUFilter, self).__init__(name = name, filter_dic = filter_dic,
                                         filter_id = filter_id)

        if filter_dic is not None:
            self.attributeName = filter_dic['attributeName']
            self.properties = filter_dic['properties']
            self.partitions = filter_dic['partitions']
        else:
            self.attributeName = attributeName
            self.partitions = partitions
            self.properties = properties

    def is_label_defined(self, label):
        """Check if a given filtering value's label is defined for this FDU

        :param label: the label to look for
        :return: True if the label was found, False otherwise
        """
        label = label.strip()
        for filtering_value in self.partitions:
            if filtering_value['name'] == label:
                return True
        return False

    def is_field_value_defined(self, field, value):
        """Check if a given filtering value as the field value for this FDU

        :param field: Field with value to look for
        :param value: Value to check if defined
        :return: True if the field value is defined. False otherwitse
        """
        if field not in self.get_fields():
            return False
        for filtering_value in self.partitions:
            if filtering_value['properties'][field] == value:
                return True
        return False

    def get_labels_for(self, field, value):
        """Get labels for filtering values including the field value.

        :param field: Field with value to look for
        :param value: Value to get labels for
        :return: Array with labels having this field value
        """
        if field not in self.get_fields():
            return []
        labels = []
        for filtering_value in self.partitions:
            if filtering_value['properties'][field] == value:
                labels.append(filtering_value['name'])
        return labels

    def get_fields(self):
        """
        :return: Fields concerned by the filter"""
        return self.attributeName

    def get_labels(self):
        """Return all labels corresponding to FDU's filtering values

        :return: List of labels or an empty list if there's no defined labels
        """
        labels_list = []
        for filtering_value in self.partitions:
            labels_list.append(filtering_value['name'])
        return labels_list

    def __str__(self):
        """
        :return: A string representation of the FDU filter
        """
        filter_str = ['FDU Filter "{}"; attribute name(s): {};'.format(self.name,
                                                                       self.attributeName),
                      "\n\tFiltering values = "]
        for filter_val in self.partitions:
            filter_str.append('label:' + filter_val['name'] + '; ')
            for attr in self.attributeName[:-1]:
                filter_str.append('values:' + filter_val['properties'][attr] + ', ')
            if len(self.attributeName) == 1:
                filter_str.append('value:')
            filter_str.append(filter_val['properties'][self.attributeName[-1]] + '; ')
        if self.id is not None:
            filter_str.append("id {}".format(self.id))
        return "".join(filter_str)



from arcgis.gis import GIS

from orionpy.orioncore.Orion import Orion
from .servicesgismanager import ServicesGISManager


class OrionGIS(Orion):
    def __init__(self, username, password, url_machine, portal="portal", verify_cert=True):
        self._orion = super().__init__(username, password, url_machine, portal, verify_cert)
        portal_url = url_machine + '/' + portal
        self._gis = GIS(portal_url, username, password, verify_cert = verify_cert)
        self._services_gis_mgr = ServicesGISManager(self._gis)

    @property
    def gis(self):
        return self._gis

    @property
    def services_gis_mgr(self):
        return self._services_gis_mgr



class BuildFilterError(ValueError):
    """Exception class to raise if an error occurs while creating a filter"""
    def foo(self):
        pass


class FilteringValuesError(ValueError):
    """Exception class to raise if an error occurs while setting filtering values"""
    def foo(self):
        pass



from .Fields import Fields
from .Resource import Resource


class Layer(Resource):
    # TODO add class LayerGroup inheriting from Layer ?
    """Class allowing to handle layers resources (components of a service)
    """

    def __init__(self, description, parent_service_url, capabilities, is_managed=False):
        """Layer constructor

        :param parent_service_url: The containing service formatted url
        :param capabilities: right capabilities on this layer
        :param is_managed: Say if rights are managed by aOB
        :param description: a dictionary containing a layer definition
        parentLayerId = -1 if no parentLayer
        subLayerIds : null if no sublayers (not a group)
        defaultVisibility (bool) : ??
        """
        capa = capabilities.split(',')
        super().__init__(parent_service_url + "/" + str(description['id']), capabilities = capa,
                         parent_access = parent_service_url, is_managed = is_managed)
        self.id = str(description['id'])
        self.name = description['name']

        # parentLayerId is used if the layer has a group-layer parent
        self.parentLayerId = description['parentLayerId']

        # This parameter is used if the layer is actually a group of layer.
        self.subLayerIds = description['subLayerIds']
        self._is_group = True if self.subLayerIds else False

        self.defaultVisibility = description['defaultVisibility']
        self.minScale = description['minScale']
        self.maxScale = description['maxScale']
        self.capabilities = capa

        self._fields = Fields(self.parent_access, self.id)

    @property
    def fields(self):
        """
        :return: List of fields corresponding to this layer
        """
        return self._fields

    # ----- Test methods -----

    def _can_activate_FDU(self, group_rights, filt, group):
        """Verify if can apply a given filter to this layer

        :param group_rights: Contains current rights on this layer
        :param filt: filter to apply
        :param group: the group for which apply the filter
        :return: True if all tests are passed. False otherwise
        """
        if not super()._can_activate_FDU(group_rights, filt, group):
            return False
        for attr in filt.get_fields():
            if self.fields.get(attr) is None:
                print('In order to apply filter on a layer, its attributes has to be in fields')
                return False
        return True

    def is_group(self):
        """Method saying if the layer is a group of layers or not"""
        return self._is_group

    def has_parent_layer(self):
        """Allows to know if a layer is in a group or not."""
        return self.parentLayerId != -1

    def has_sub_layers(self):
        """Allows to know if a layer has sub layers"""
        return self.subLayerIds is not None

    # ----- Access methods -----

    def get_sub_layers_ids(self):
        """If this is a group of layers, returns list of its sublayers"""
        if self.has_sub_layers():
            return self.subLayerIds
        print('[WARNING] Layer {}({}) not a group of layers. or has no sub-layers.'
              'None is returned'.format(self.name,
                                        self.id))
        return None

    def get_parent_layer_id(self):
        """Returns the parent layer's id (if there is one)"""
        if self.parentLayerId == -1:
            print('[WARNING] no parent layer. None is returned')
            return None
        return self.parentLayerId

    @staticmethod
    def get_type():  # TODO mention LAYER-GROUP here ?
        return "LAYER"

    def __str__(self):
        layer_type = 'group of layers' if self.is_group() else 'simple layer'
        layer_str = [
            'Layer {}; name {}; type {}; for Service {}.'.format(self.id, self.name, layer_type,
                                                                self.parent_access)]
        if self.is_group() and self.has_sub_layers():
            layer_str.append("\n\tIts sub-layers'IDs are : {}".format(self.get_sub_layers_ids()))

        return "".join(layer_str)



# https://integration.arcopole.fr/Orion/orion/admin/tree/object/BUSINESS/Cadastre

from .BusinessResource import BusinessResource

# TODO : factorize with Resource !
class StatsResource(BusinessResource):
    def __init__(self, resource_description):
        """

        :param resource_description: Description of the stat resource
        """
        super().__init__(resource_description)






from orionpy.orioncsv import orioncsv
import json


class CSVFilteringCadastre(orioncsv.OrionCSV):
    def __init__(self, orion):
        super().__init__(orion)
        self.orion = orion
        self.group = "group"
        self.right = "role"
        self.filter_values = "scope"
        self.group_i = 0
        self.right_i = 1
        self.filter_values_i = 2

    def generate(self, csv_path, json_path):
        """Generate a csv file with the groups, the rights and
        the filter values, ready to import

        :param csv_path: path of the csv file you want to generate
        :param json_path: path of the json file where the data is stored
        """
        self._csv_process(csv_path, self._json_extract_information(json_path))

    def _init_generate(self, json_path):
        """Generate a dictionnary from

        :param json_path: path of the json file where the data is stored
        :return: Dictionary of needed values to update the rights
        """
        with open(json_path, 'r', encoding="utf-8") as json_data:
            data_list_input = json.load(json_data)
            data_dict_input = data_list_input[0]
        profiles_dict = data_dict_input["profiles"]
        return profiles_dict

    def _json_extract_information(self, json_path):
        """Extract the needed informations from the json file

        :param json_path: path of the json file where the data is stored
        :return: A list of dict with json needed informations included. List : [group, role, scope].
        """
        data_list_output = list()
        profiles_dict = self._init_generate(json_path)

        for dictionary in profiles_dict:
            temp_dict = dict()
            if self.group in dictionary.keys():
                temp_dict[self.group] = dictionary[self.group]
            if "rules" in dictionary.keys():
                sub_set = set()
                for sub_dico in dictionary["rules"]:
                    self._add_right_or_filter_values(sub_dico, temp_dict, sub_set)
                temp_dict[self.filter_values] = sub_set
                if self.right not in temp_dict.keys():
                    temp_dict[self.right] = "public"  # if there is no "right" defined, public
            data_list_output.append(temp_dict)
        return data_list_output

    def _add_right_or_filter_values(self, sub_dico, temp_dict, sub_set):
        """Add right or filter values

        :param sub_dico:
        :param temp_dict:
        :param sub_set:
        """
        if self.right in sub_dico.keys():
            temp_dict[self.right] = sub_dico[self.right]
        elif self.filter_values in sub_dico.keys():
            sub_set.add(sub_dico[self.filter_values])

    def _csv_process(self, csv_path, data_list):
        """Create the CSV with the data list

        :param csv_path: path of the csv file you want to generate
        :param data_list: json informations extracted in a list of dict
        """
        csvfile = open(csv_path, 'w', newline='')
        first_line = [self.group, self.right, self.filter_values]
        writer = self.get_csv_writer(csvfile, first_line)
        writer.writeheader()
        for dico in data_list:
            group = dico[self.group]
            role = dico[self.right]
            if role == set():
                role = ''
            scope = ''
            if scope != set():
                for element in dico[self.filter_values]:
                    scope += str(element) + ','
                scope = scope[:-1]
            writer.writerow({self.group: group, self.right: role, self.filter_values: scope})
        csvfile.close()

    def read_and_apply(self, csv_path):
        """Parse the csv file and apply the rights on aOB

        :param csv_path: path of the csv file
        """
        cadastre_resource = self.orion.services.get_cadastre_resource()
        fdu_filter = self.orion.filters.get_with_id(cadastre_resource.associated_filter_id)
        cadastre_resource.init_filter_access(fdu_filter)
        csv_file = open(csv_path, 'r', encoding="utf-8")
        csv_reader = self.get_csv_reader(csv_file)
        field = "id_comm"
        next(csv_reader)
        line_count = 1
        for row in csv_reader:
            print("\n", row,"\n")
            line_count += 1
            group = self.orion.groups.get(row[self.group_i])
            if group == None:
                print("The line ", line_count, " ( group : \"",row[self.group_i] ,"\" ) of the CSV will not be used, due "
                        "to an error. Please check the group name in the CSV or check if the group is define on aOB")
                continue
            filter_values = list(filter(lambda value: value != "", row[self.filter_values_i].split(','))) # make a list of filter values
            self._set_filter_values(group, fdu_filter, filter_values, field, row[self.right_i])
            if not cadastre_resource.update_right_from_string(group, row[self.right_i]):
                print("The line ", line_count, " ( group : \"",row[self.group_i] ,"\" ) of the CSV will not be used, due "
                                                                    "to an error. Please check the right in the CSV.")
                continue
            self._activate_or_deactivate_filter(row, group, filter_values, cadastre_resource)

    def _activate_or_deactivate_filter(self, row, group, filter_values, cadastre_resource):
        """Determine if the filter needs to be activated / deactivated

        :param row:
        :param group:
        :param filter_values:
        :param cadastre_resource:
        """
        if row[self.right_i] == "public" and filter_values:
            cadastre_resource.activate_filter(group)
        elif row[self.right_i] == "public" and not filter_values:
            cadastre_resource.deactivate_filter(group)

    def _set_filter_values(self, group, fdu_filter, filter_values, field, right_string):
        """Set the filter values / particular case with no values and nominative access

        :param group:
        :param fdu_filter:
        :param filter_values:
        :param field:
        :param right_string:
        """
        if right_string == 'nominatif' and not filter_values:
            group.set_filter_values(fdu_filter, add_all_values=True)
        else:
            group.set_filter_field_values(fdu_filter, filter_values, field)



import json


class Item:

    def __init__(self, data):
        self.data = data

    def get_id(self):
        """
        Return the id of the item
        :return: str id of the item
        """
        return self.data["id"]

    def get_owner(self):
        """
        Return the owner of the item
        :return: str owner of the item
        """
        return self.data["owner"]

    def get(self, element):
        """
        Return the element of the item
        :return: element of the item
        """
        return self.data[element]

    def __str__(self):
        return json.dumps(self.data, indent=4, ensure_ascii=False)

    def __repr__(self):
        return str(self)



import requests
import urllib3
from sys import exit


class _Singleton(type):
    """
    metaclass of Orion.
    type is the base metaclass
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        """

        :param args: list of arguments
        :param kwargs: Dictionary [keyword arguments]= values of these arguments
        :return:
        """
        # If the class was not created yet (not in our list)...
        if cls not in cls._instances:
            # call the __call__ method of type. So, calls the class constructor
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class RequestManager(metaclass = _Singleton):
    """Class managing communication with the REST API.
    The only one to use the requests python library.
    """

    def __init__(self, output_format='json', token=None, verify=True):
        self._basedata = {'f': output_format}
        if not verify:
            print('[WARNING] Unverified HTTPS request is being made. '
                  'Adding certificate verification is strongly advised.')
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        self.verify = verify
        if token is not None:
            self._basedata['token'] = token

    def set_token(self, token):
        """Set a given token to be used in all other requests.

        :param token: token to set. Must be valid
        """
        self._basedata['token'] = token

    def get_basedata(self):
        """

        :return: basedata used for requests (token and format)
        """
        return self._basedata

    def _merge_data(self, params, keep_base_parameters):
        """Adds new parameters to the default ones"""
        if params is None and keep_base_parameters:
            params = self._basedata
        elif params is not None and keep_base_parameters:
            # merge the two dictionaries in one
            params = {**self._basedata, **params}
        return params

    def get(self, url, params=None, keep_base_parameters=True):
        """Does a GET request on the url

        :param url: where to do the requests
        :param params: parameters for the request
        :param keep_base_parameters: Boolean if want to keep default parameters.
        :return: the request's answer.
        """
        params = self._merge_data(params, keep_base_parameters)

        try:
            req = requests.get(url, params = params, verify = self.verify)
            if req.status_code != requests.codes.ok:
                req.raise_for_status()
        except requests.exceptions.HTTPError as he:
            print('HTTP error : ', he)
            exit(-1)
        return req

    def post(self, url, data=None, keep_base_parameters=True):
        """Does a POST request on the url

        :param url: where to do the requests
        :param data: parameters for the request
        :param keep_base_parameters: Boolean if want to keep default parameters.
        :return: the request's answer
        """
        data = self._merge_data(data, keep_base_parameters)
        try:
            req = requests.post(url, data = data, verify = self.verify)
            if req.status_code != requests.codes.ok:
                req.raise_for_status()
        except requests.exceptions.ConnectionError as ce:
            print('Connection error :', ce)
            print('url :', url)
            exit(-1)
        except requests.exceptions.HTTPError as he:
            print('HTTP error : ', he)
            exit(-1)
        return req

    def post_in_python(self, url, data=None, keep_base_parameters=True):
        """Does a POST request on the url and returns the answer as a Python structure

        :param url: where to do the requests
        :param data: parameters for the request
        :param keep_base_parameters: Boolean if want to keep default parameters
        :return: Request answer converted from json to python
        """
        struct = self.post(url, data, keep_base_parameters).json()
        if 'error' in struct:
            print('ERROR in response : ', struct['error'])
            print('url : ', url)
            exit(-1)
        return struct

    def get_python_answer(self, response):
        """Convert a string in json format to a python dictionary

        :param response: Answer of http request.
        :return: the json string converted in a python dictionary
        """
        struct = response.json()
        if 'error' in struct:
            print('ERROR in response : ', struct['error'])
            print('url : ', response.url)
            exit(-1)
        return struct

    def get_in_python(self, url, params=None, keep_base_parameters=True):
        """Does a get request and return the answer formatted as a Python structure

        :return: Request answer converted from json to python
        """
        # TODO use it directly instead of get_python_answer
        return self.get_python_answer(self.get(url, params, keep_base_parameters))



from .Fields import Fields
from .Resource import Resource


class Table(Resource):
    def __init__(self, description, parent_service_url, is_managed=False):
        super().__init__(parent_service_url + "/" + str(description['id']),
                         parent_access = parent_service_url, is_managed = is_managed)
        self.id = str(description['id'])
        self.name = description['name']

        self._fields = Fields(self.parent_access, self.id)

    @property
    def fields(self):
        """
        :return: List of fields corresponding to this layer
        """
        return self._fields

    def _can_activate_FDU(self, group_rights, filt, group):
        """Verify if can apply a given filter to this table
        Only possible if filter's attributes are in table's fields.

        :param group_rights: Contains current rights on this table
        :param filt: filter to apply
        :param group: the group for which apply the filter
        :return: True if all tests are passed. False otherwise
        """
        if not super()._can_activate_FDU(group_rights, filt, group):
            return False
        for attr in filt.get_fields():
            if self.fields.get(attr) is None:
                print('In order to apply filter on a table, its attributes has to be in fields')
                return False
        return True

    @staticmethod
    def get_type():
        return "TABLE"

    def __str__(self):
        return 'Table {}; name {}; for Service {}'.format(self.id, self.name, self.parent_access)



from orionpy.orioncore.Filter import _FDUFilter
from .orioncsv import OrionCSV


class CSVFilteringValues(OrionCSV):
    # TODO handle labels starting with a 0 !
    def __init__(self, orion):
        super().__init__(orion)
        self.filter_head = 'Filter'
        self.label_head = 'Label'
        self.filter_i = 0
        self.label_i = 1
        self.group_i = 2

    def generate(self, csv_path):
        """Create/Edit a csv file with a summary of all FDU and filtering set for all groups

        :param csv_path: path to the csv files
        """
        print('Generating csv file with a summary of all FDU and filtering values set...')
        # Prepares data and header required
        header = [self.filter_head, self.label_head]
        groups = self.orion.groups.all()
        group_names = []
        for group in groups:
            group_name = group.get_name()
            header.append(group_name)
            group_names.append(group_name)
        filters = self.orion.filters.all_fdu()

        group_labels = dict.fromkeys(group_names, {})

        # TODO factorize following lines with CSVRightsManagement
        try:
            with open(csv_path, "w+", newline = '') as csv_file:
                writer = self.get_csv_writer(csv_file, header)
                writer.writeheader()
                row = {}
                for fdu_filter in filters:
                    # Get the data that interest us (= filtering values and if they're activated)
                    print('\tGetting filtering values for filter "{f}"...'.format(f=fdu_filter.get_name()))
                    for group in groups:
                        group_labels[group.get_name()] = group.get_activated_labels(fdu_filter)
                    # Fills row and write in csv.
                    row[self.filter_head] = fdu_filter.get_name()
                    for label in fdu_filter.get_labels():
                        row[self.label_head] = label
                        for group_name, labels_on in group_labels.items():
                            row[group_name] = '1' if labels_on[label] else '0'
                        writer.writerow(row)
            print('Done ! Csv file containing all filtering values was generated at', csv_path)
        except IOError as e:
            print(e)

    def read_and_apply(self, csv_path):
        """Parse a csv file with fdu, groups and filtering values.
        Does so in order to associate the filtering values in Orion

        :param csv_path: path to the csv files
        """
        print('Scanning file', csv_path, '....')

        fdu_filter = None
        # TODO Call methods using requests (call set_filter_values ?) at the end of the loop
        try:
            with open(csv_path, "r") as csv_file:
                reader = self.get_csv_reader(csv_file)
                # Prepares data
                groups = self._get_groups(header = next(reader), start = self.group_i)
                if not groups:
                    print('None of the group exists..')
                    return

                for row in reader:
                    # If the filter wasn't set or is different from the previous one...
                    if fdu_filter is None or fdu_filter.get_name() != row[self.filter_i]:
                        fdu_filter = self.orion.filters.get(row[self.filter_i])
                        if fdu_filter is None or not isinstance(fdu_filter, _FDUFilter):
                            print('Filter', row[self.filter_i], 'does not exist or not a FDU')
                            continue  # goes to the next line.
                        print('\tAnalyzing filtering values for "{}"...'.format(fdu_filter.get_name()))

                    label = row[self.label_i]
                    if not fdu_filter.is_label_defined(label):
                        print('Label', label, 'not in filter')
                        continue
                    for add_filterval, group in zip(row[self.group_i:], groups):
                        if group is None:  # The group doesn't exist
                            continue
                        # TODO call this at the end (modification for 1 filter of >1 labels)
                        self._change_group_filtering_value(group, fdu_filter, label, add_filterval)
            print('Done ! All possible filtering values were successfully modified !')
        except IOError as e:
            print(e)

    def _change_group_filtering_value(self, group, fdu_filter, label, add_filterval):
        """For a group, add or remove a filtering value based on add_filterval's value"""
        # TODO : just call 'set_filter_values
        if int(add_filterval):  # if the value is different of 0, we must set filtering value
            group.add_filtering_values(fdu_filter, [label])
        else:
            group.remove_filtering_values(fdu_filter, [label])



from orionpy.orioncore.resources.Services import Services
from orionpy.orioncore import cfg_global


# https://front.arcopole.fr/Orion/orion/admin/tree/object/SERVICES/Cannes/EspacesVerts_FeatureServer/__refreshPortalItemSharingData
# GET
# f & token

class ServicesGISManager(Services):
    def __init__(self, gis):
        super().__init__()
        self._gis = gis

    def get_gis(self, element_id):
        service = super().get(element_id)
        return self._gis.content.get(service.id)  # arcgis

    def share(self, service, group_id):
        """Shares a given service with a given group"""
        if not cfg_global.is_federated:
            print('Method not available if ArcGIS not federated')
            return
        if self.is_shared_with(service, group_id):
            print('Service', service.get_access_url(), 'already shared with group', group_id)
            return
        service_arcgis = self._gis.content.get(service.id)
        print(service_arcgis.share(groups = group_id))

    def unshare(self, service, group_id):
        """Unshares a given service with a given group"""
        if not cfg_global.is_federated:
            print('Method not available if ArcGIS not federated')
            return
        if not self.is_shared_with(service, group_id):
            print('Service', service.get_access_url(), 'already not shared with group', group_id)
            return
        service_arcgis = self._gis.content.get(service.id)
        print(service_arcgis.unshare(groups = group_id))

    def get_groups_id_shared(self, service):
        """Get the list of groups'ids where service is shared with"""
        if not cfg_global.is_federated:
            print('Method not available if ArcGIS not federated')
            return
        service_arcgis = self._gis.content.get(service.id)
        url = self.url_manager.resource_sharing_url(service_arcgis.id,
                                                    service_arcgis.owner)
        req = self.request.post_in_python(url)
        groups_ids = req['sharing']['groups']
        if req['sharing']['access'] == 'org':
            groups_ids.append('org')
        elif req['sharing']['access'] == 'public':
            groups_ids.append('public')
            groups_ids.append('org')

        return groups_ids

    def is_shared_with(self, service, group_id):
        """Check if a given group is shared with a service"""
        if not cfg_global.is_federated:
            print('Method not available if ArcGIS not federated')
            return
        groups_ids = self.get_groups_id_shared(service)
        for group_ids in groups_ids:
            if group_id == group_ids:
                return True
        return False



# ---- Get informations about a group----
# https://{machine}/{api}/rightmanagement/groups/{group}
# GET
# f & token

# ---- Get rights on resources ----
# https://{machine}/{api}/rightmanagement/groups/{group}/authorizedResources

# ---- Get rights on services ----
# {getRightsResources_URL}/SERVICES

# ---- Get right on a particular service ----
# {getRightServices_URL}/{service}

# ---- Get rights on a layer ----
# {getRightsParticularService_URL}/{layer}

# ---- Get rights on a field ----
# {getRightsLayer_URL}/{field}

# ---- Set filtering values ----
# {groupInformationURL}/__configure
# value = {"name":$GROUP_ID$,
#          "domaine":$GROUP_DOMAIN$,
#          "perimeters":
#               [{"dimension":$FILTER_ID$,
#                 "valeures":$ONE_SELECTED_FILTERING_VAL$
#                 }*
#                ]
#         }
# https://{machine}/{api}/rightmanagement/groups/{group}
#    {
#     "title":"Organisation",
#     "name":"org",
#     "domain":"default",
#     "isSuperAdmin":false,
#     "perimeters": # defined filtering values
#      [
#      ],
#     "properties":
#      {
#       "builtinRole":"authenticated"
#      }
#    }

# "name": "128703724fce49358406f2dfcf8d43aa",
# "nodeType": "GroupNode",
# "title": "Accueil"

import json

from .Subject import Subject


class Group(Subject):
    """Single group-data handling class"""

    def __init__(self, title, group_id, profile_id = None):
        super().__init__(title, group_id, "groups", profile_id)

    def is_system_subject(self):
        return self.id == 'org' or self.id == 'public'


# https://front1051.arcopole.fr/Orion/orion/admin/tree/object/SERVICES/Cannes/EspacesVerts_FeatureServer/2
# {
# "nodeInformationData":"LAYER",
# "definition":
#  {
#   "id":$layer_id$,
#   "name":$layer_name$,
#
#   "fields":
#    [
#       {
#        "name":$field_name/id$,
#        "type":$field_type$,
#        "alias":$field_alias$,
#        "editable":false,
#        "nullable":false,
#         [length]: $len$,
#       },
#    ],
#   "restRelativeURL":"Cannes/EspacesVerts/FeatureServer/2",
#  }
# }

from enum import Enum

from .Field import Field
from ..Elements import Elements


class FieldsKeys(Enum):
    fields_key = 'fields'
    def_key = 'definition'


class Fields(Elements):
    """Class allowing to get access to a list of fields for a particular layer
    """

    # TODO : Fields can also be identified with there alias. Be able to get a field using this alias.
    # ==> Add a method get_with_alias

    def __init__(self, service_url, layer_id):
        super().__init__()
        self.parent_url = service_url
        self.layer_id = layer_id
        self.parent_access = service_url + "/" + layer_id

    def _update(self):
        """Make sure that list of fields is up to date"""
        self._elements = {}
        parent_def = self.request.get_in_python(
            self.url_manager.resource_definition_url(self.parent_access))
        if parent_def[FieldsKeys.def_key.value][FieldsKeys.fields_key.value] is None:
            return
        for field in parent_def[FieldsKeys.def_key.value][FieldsKeys.fields_key.value]:
            if 'isManaged' in parent_def[FieldsKeys.def_key.value]:
                is_managed = parent_def[FieldsKeys.def_key.value]['isManaged']
            else:
                is_managed = False
            self._elements[field['name']] = Field(field, self.parent_url, self.layer_id,
                                                  parent_def[FieldsKeys.def_key.value]['capabilities'],
                                                  is_managed)



from ..Elements import Elements
from .Table import Table


class Tables(Elements):
    def __init__(self, parent_service_url):
        """Initialize and build our list of tables

        :param parent_service_url: url (unique) for the service containing all the tables
        """
        super().__init__()

        self.service_url = parent_service_url

    def _update(self):
        """Update list to be consistent with database"""
        self._elements = {}
        service_def_url = self.url_manager.resource_definition_url(self.service_url)
        service_def = self.request.get_in_python(service_def_url)
        for table in service_def['definition']['tables']:
            if 'isManaged' in service_def['definition']:
                is_managed = service_def['definition']['isManaged']
            else:
                is_managed = False
            self._elements[table['name']] = Table(table, self.service_url,
                                                  is_managed)

    def get_with_id(self, table_id):
        """Returns a table given its id
        """
        self._update()
        table_id = table_id.strip()
        for key, table in self._elements.items():
            if table.id == table_id:
                return table
        print("[WARNING] id", table_id, "doesn't exist, None is returned")
        return None



# ---- Get informations about a subject----
# https://{machine}/{api}/rightmanagement/groups/{subject}
# GET
# f & token

# ---- Get rights on resources ----
# https://{machine}/{api}/rightmanagement/groups/{subject}/authorizedResources

# ---- Get rights on services ----
# {getRightsResources_URL}/SERVICES

# ---- Get right on a particular service ----
# {getRightServices_URL}/{service}

# ---- Get rights on a layer ----
# {getRightsParticularService_URL}/{layer}

# ---- Get rights on a field ----
# {getRightsLayer_URL}/{field}

# ---- Set filtering values ----
# {groupInformationURL}/__configure
# value = {"name":$GROUP_ID$,
#          "domaine":$GROUP_DOMAIN$,
#          "perimeters":
#               [{"dimension":$FILTER_ID$,
#                 "valeures":$ONE_SELECTED_FILTERING_VAL$
#                 }*
#                ]
#         }
# https://{machine}/{api}/rightmanagement/groups/{subject}
#    {
#     "title":"Organisation",
#     "name":"org",
#     "domain":"default",
#     "isSuperAdmin":false,
#     "perimeters": # defined filtering values
#      [
#      ],
#     "properties":
#      {
#       "builtinRole":"authenticated"
#      }
#    }

# "name": "128703724fce49358406f2dfcf8d43aa",
# "nodeType": "GroupNode",
# "title": "Accueil"

import json

from .Exceptions import FilteringValuesError
from .Filter import _FDUFilter
from .RequestManager import RequestManager
from .UrlBuilder import UrlBuilder


class Subject:
    # TODO way too much method to update filtering values
    """Single subject-data handling class"""

    def __init__(self, title, id, subject_type, profile_id = None):
        self.title = title
        # NB : for 'org' and 'public' groups, id is the builtinrole.
        self.id = id
        self.subject_type = subject_type
        if profile_id is not None:
            self._profile_id = profile_id
        self._url_builder = UrlBuilder()
        self._request_mgr = RequestManager()
        self.perimeters_key = 'perimeters'
        self.filter_val_key = 'valeures'
        self.filter_id_key = 'dimension'

    # ----- Updating filtering values -----

    @property
    def profile_id(self):
        if self.is_system_subject():
            return self._profile_id
        return self.id

    def set_several_filter_values(self, filters_and_values):
        """Update/Add filtering values for several FDU filters on this subject
        For other filters, keep the values already defined.

        :param filters_and_values: A dictionary : [filter] = corresponding filtering values.
        """
        for filt, filtering_values in filters_and_values.items():
            # TODO : improve it to send only one request.
            self.set_filter_values(filt, filtering_values)

    # TODO : deprecates it. Create new method 'empty_filtering_values()'
    def reset_all_filters_values(self, filters_and_values):
        """Re-set filtering values for several FDU filters (with only one request)
        If there is existing filtering values for this subject, removes it.

        :param filters_and_values: A dictionary : [filter] = corresponding filtering values.
        """
        subject_information = self._request_mgr.get_in_python(
            self._url_builder.subject_information_url(self.subject_type, self.id))
        # as we will redefine filter values, erase previous ones
        subject_information[self.perimeters_key] = []
        for filt, filtering_values in filters_and_values.items():
            if not isinstance(filt, _FDUFilter):
                print('[ERROR] Only possible to defined filtering values on a FDU filter')
                continue
            if filtering_values:
                subject_information[self.perimeters_key].append(
                    self._build_perimeters(filt, filtering_values))
            else:
                print('[ERROR] : selected_filtering_values must not be empty')

        self._apply_configuration(subject_information)

    # TODO : allow 'labels' to be values as well ?
    def set_filter_values(self, fdu_filter, labels=None,
                          add_all_values=False):
        """Update filtering values for one FDU filter on this subject

        :param fdu_filter: The FDU filter for which set filtering values
        :param labels: list of filtering values selected
        :param add_all_values: boolean to specify if want to add each value
        :raises FilteringValuesError: if problem in parameters
        """
        try:
            if not isinstance(fdu_filter, _FDUFilter):
                raise FilteringValuesError('[ERROR] Can only define filtering values on a FDU')
            # Check that the good arguments are given
            if not add_all_values:
                if labels is None:
                    raise FilteringValuesError('[ERROR] If add_all False, labels must be provided.')
                elif not labels:
                    raise FilteringValuesError('[ERROR] "labels" can\'t be an empty list')
        except FilteringValuesError as e:
            print(e.args)
            return

        if add_all_values:
            labels = fdu_filter.get_labels()

        # Get current filtering values defined for this subject
        subject_information = self._subject_information_minus_filter(fdu_filter)
        # add values for our filter
        subject_information[self.perimeters_key].append(
            self._build_perimeters(fdu_filter, labels))
        # Applies change and print result ({type:configue} if ok)
        self._apply_configuration(subject_information)

    def add_filtering_values(self, fdu_filter, labels):
        """Adds filtering values to a subject if not already set

        :param fdu_filter: the filter for which set filtering values
        :param labels: a list containing labels to add in filtering values
        """
        subject_info = self.get_subject_informations()
        f_index = self._get_filter_index(fdu_filter.get_id(),
                                         subject_info[self.perimeters_key])
        if f_index is None:  # Nothing was found, need to create strcture for filter
            self.set_filter_values(fdu_filter, labels)
            return

        # Get filtering values already defined on this filter.
        filter_values = subject_info[self.perimeters_key][f_index][self.filter_val_key].split(',')

        something_to_add = False
        for label in labels:  # For the labels we want to add...
            # If the label isn't already defined, does it
            if label not in filter_values:
                something_to_add = True
                filter_values.append(label)

        if something_to_add:
            # Applies the new configuration
            self._apply_filtering_values(subject_info, f_index, filter_values)

    def remove_filtering_values(self, fdu_filter, labels):
        """Removed filtering values to a subject if already set

        :param fdu_filter: the filter for which remove filtering values
        :param labels: a list containing labels to remove in filtering values
        """
        subject_info = self.get_subject_informations()
        f_index = self._get_filter_index(fdu_filter.get_id(),
                                         subject_info[self.perimeters_key])
        if f_index is None:  # Nothing was found, nothing to change !
            return

        # Get filtering values already defined on this filter.
        filter_values = subject_info[self.perimeters_key][f_index][self.filter_val_key].split(',')
        if not filter_values:  # If no filtering value, nothing to remove !
            return

        something_to_remove = False
        for label in labels:  # For the labels we want to delete...
            # If label is defined, remove it
            if label in filter_values:
                something_to_remove = True
                filter_values.remove(label)

        if something_to_remove:
            # Applies the new configuration
            self._apply_filtering_values(subject_info, f_index, filter_values)

    def set_filter_field_values(self, fdu_filter, values, field):
        """Update filtering values for one FDU filter on this subject.
        Instead of label, use a field values

        :param fdu_filter: The FDU filter for which set filtering values
        :param values: list of filtering values selected
        :param field: field with values
        :raises FilteringValuesError: if problem in parameters
        """
        try:
            if not isinstance(fdu_filter, _FDUFilter):
                raise FilteringValuesError('[ERROR] Can only define filtering values on a FDU')
            if values is None or not values:
                raise FilteringValuesError('[ERROR] "values" can\'t be an empty list')
            if not field:
                raise FilteringValuesError('[ERROR] "field" must be provided')
        except FilteringValuesError as e:
            print(e.args)
            return

        # Get current filtering values defined for this subject
        subject_information = self._subject_information_minus_filter(fdu_filter)
        # add values for our filter
        subject_information[self.perimeters_key].append(
            self._build_perimeters(fdu_filter, values, field))

        # Applies change and print result ({type:configue} if ok)
        self._apply_configuration(subject_information)

    def _apply_filtering_values(self, subject_info, filter_index, filter_values):
        """Applies modification of filtering values for a filter on a subject."""
        subject_info[self.perimeters_key][filter_index][self.filter_val_key] = ','.join(
            filter_values)
        self._apply_configuration(subject_info)

    def _get_filter_index(self, filter_id, filtering_values):
        """Search if a filter has defined values in a subject. If so, returns its index

        :param filter_id: filter to look for
        :param filtering_values: list of defined filtering values for this subject
        :return: filter index if found. None otherwise
        """
        return next((index for (index, filt) in enumerate(filtering_values)
                     if filt[self.filter_id_key] == filter_id), None)

    def _get_filter_values(self, filter_id):
        """Returns filtering values defined for this subject on a given filter"""
        info_url = self._url_builder.subject_information_url(self.subject_type, self.id)
        subject_info = self._request_mgr.get_in_python(info_url)
        for perim in subject_info[self.perimeters_key]:
            if perim[self.filter_id_key] == filter_id:
                return perim[self.filter_val_key]
        return None

    def _apply_configuration(self, config):
        """Applies a pre-defined configuration to our subject

        :param config: the configuration required.
        """
        self._request_mgr.get(self._url_builder.subject_configure_url(self.subject_type, self.id),
                              params = {'value': json.dumps(config)})
        print('Successfully applied change on configuration for subject "{}" !'.format(self.title))

    def _subject_information_minus_filter(self, filter_to_set):
        """Get the current subject configuration, removing the selected filter values

        :param filter_to_set: filter for which remove filtering values if it's defined
        :return: A dictionary representing the formatted subject configuration
        """
        subject_information = self.get_subject_informations()
        self._remove_current_filter_value(subject_information[self.perimeters_key], filter_to_set.id)
        return subject_information

    def _build_perimeters(self, filter_concerned, filtering_values, field=""):
        """Builds the associtation filter-filtering values. Formatted as required by the API.

        :param filter_concerned: the filter for which associate filtering values
        :param filtering_values: the filtering values to add to this filter
        :param field: (optional) corresponding field if filtering_values are field value
        :return: association filter-values as : {'dimension': filter_id, 'valeures':"val1, val2..."
        """
        approved_values = []
        # Handle if there are values in selected_filtering_values which aren't set for filter
        for filtering_value in filtering_values:
            if filter_concerned.is_label_defined(filtering_value):
                approved_values.append(filtering_value)
            elif field and filter_concerned.is_field_value_defined(field, filtering_value):
                approved_values += filter_concerned.get_labels_for(field, filtering_value)
            else:
                print('Value', filtering_value, 'not defined for this filter. It will not be set')
        return {self.filter_id_key: filter_concerned.id,
                self.filter_val_key: ','.join(approved_values)}

    def _remove_current_filter_value(self, perimeters, filter_id):
        """If filtering values were already defined for a selected filter, removes it

        :param perimeters: the list of filter values to look in
        :param filter_id: the filter to look for (and remove if found)
        """
        # TODO find a better name
        # TODO use index(filter_id) and pop(index) methods.
        for filtering_values in perimeters:
            if filtering_values[self.filter_id_key] == filter_id:
                perimeters.remove(filtering_values)
                return
        print('Defining filtering values for', filter_id)

    # ----- Test methods -----

    def has_defined_filter_values(self, fdu_filter):
        """Check if the subject has filtering values defined for a specified filter

        :param fdu_filter: filter to check for
        :return: True if subject has defined filtering values, False otherwise
        """
        info_url = self._url_builder.subject_information_url(self.subject_type, self.id)
        subject_info = self._request_mgr.get_in_python(info_url)
        for perim in subject_info[self.perimeters_key]:
            if perim[self.filter_id_key] == fdu_filter.id:
                return True
        return False

    def has_label_set_for_filter(self, fdu_filter, label):
        """Check if a specific label is enabled as filtering values for this subject

        :return: True if this label is activated. False otherwise."""
        filter_values = self._get_filter_values(fdu_filter.get_id())
        if filter_values is None:
            return False
        return label in filter_values

    # TODO : rename method
    def is_system_subject(self):
        return False

    # ----- Access methods -----
    def get_name(self):
        """

        :return: The name of this subject
        """
        # TODO use it everywhere
        return self.title

    def get_id(self):
        """

        :return: Id of this subject
        """
        # TODO use it everywhere
        return self.id

    def get_activated_labels(self, fdu_filter):
        """Returns a structure saying which labels are activated for a filter on this subject

        :param fdu_filter: Filter analyzed
        :return: Structure with label key and value = True if label set, False otherwise
        """
        # Builds a dictionary with labels as keys and False as default value
        labels_activated = dict.fromkeys(fdu_filter.get_labels(), False)
        filter_values = self._get_filter_values(fdu_filter.get_id())
        if filter_values is None:  # If no filtering values was set on this filter
            return labels_activated
        for label in labels_activated.keys():
            if label in filter_values:
                labels_activated[label] = True
        return labels_activated

    def get_subject_informations(self):
        subject_information_url = self._url_builder.subject_information_url(self.subject_type, self.id)
        return self._request_mgr.get_in_python(subject_information_url)

    def print_defined_filtering_values(self):
        """Prints a list of defined filtering values in a subject"""
        subject_information = self._request_mgr.get_in_python(
            self._url_builder.subject_information_url(self.subject_type, self.id))
        print('The filtering values set for the subject', self.get_name())
        # as we will redefine filter values, erase previous ones
        filtering_str = []
        for val in subject_information[self.perimeters_key]:
            filtering_str.append("\tOn filter {}, "
                                 "value(s) defined is/are : {}\n".format(val[self.filter_id_key],
                                                                         val[self.filter_val_key]))
        print("".join(filtering_str))

    def __str__(self):
        type_str = None
        if self.subject_type == "groups":
            type_str = "Group"
        elif self.subject_type == "users":
            type_str = "User"
        return '{} "{}"; id {}'.format(type_str, self.title, self.id)



# https://integration.arcopole.fr/Orion/orion/admin/tree/object/BUSINESS/Cadastre

import abc

import json
from .Resource import RightLevel
from ..RequestManager import RequestManager
from ..UrlBuilder import UrlBuilder


# TODO : factorize with Resource !
class BusinessResource(metaclass = abc.ABCMeta):
    def __init__(self, resource_description):
        """

        :param resource_description: Description json of resource
        """
        self.id = resource_description.get('_id') or resource_description.get('id') or resource_description.get('name')
        self._url_builder = UrlBuilder()
        self._request_mgr = RequestManager()
        self._description = resource_description

    def __str__(self):
        """Provides a string representation of a cadastre resource"""
        resource_str = 'Resource id : {}'.format(self.id)
        return resource_str



import json
from enum import Enum

from .Elements import Elements
from .Exceptions import BuildFilterError
from .Filter import FilterFactory, _FDUFilter


class FiltersKeys(Enum):
    name_key = "name"
    properties_key = "properties"
    dimension_key = "dimensions"


class Filters(Elements):
    """Class allowing to get access to the list of defined filters
    """

    def __init__(self):
        """
        Initialize and build our list of filters
        """
        super().__init__()
        # Keeps the filters as dictionary (useful when update DB)
        self._filters_as_dic = []

    def _update(self):
        """Update the list of filter to make sure it's consistent with our database.
        """
        filters_description = self.request.get_in_python(self.url_manager.filter_list_url())
        self._elements = {}
        self._filters_as_dic = []
        for filter_description in filters_description[FiltersKeys.dimension_key.value]:
            filt = FilterFactory.build_filter(filter_dic = filter_description)
            if filt is not None:
                self._elements[filt.name] = filt
                self._filters_as_dic.append(filter_description)
            else:
                print('ERROR in filter argument')

    def get_with_id(self, filter_id):
        """Look for a particular group in the list using its id

        :param filter_id: id of the group to look for
        :return: the required group or None if nothing found
        """
        self._update()
        filter_id = filter_id.strip()
        for key, filter_description in self._elements.items():
            if filter_description.id == filter_id:
                return filter_description
        return None

    def all_fdu(self):
        """
        :return: the list of FDU filters' values
        """

        self._update()
        fdu_filters = {}
        for key, filt in self._elements.items():
            if isinstance(filt, _FDUFilter):
                fdu_filters[key] = filt
        return fdu_filters.values()

    @staticmethod
    def _label_already_used(filtering_values, label):
        """Check if a given label is already used
        In the list of filtering values, a label can only be used once

        :return: True if label found. False otherwise
        """
        for element in filtering_values:
            if element[FiltersKeys.name_key.value] == label:
                return True
        return False

    @staticmethod
    def _build_partition_1_attr(field, filtering_values):
        """Builds the structure for the list of filtering_values ('partitions') for 1 attributes

        :param field: Name of the field concerned
        :param filtering_values: list of values where to do the filtering
        :return: The list of filtering values (partitions) created
        """
        partitions = []
        for filtering_val in filtering_values:
            if Filters._label_already_used(partitions, filtering_val):
                print('[ERR FILTERING_VALUES] label', filtering_val, 'already in list so not added')
                continue
            partitions.append({FiltersKeys.name_key.value: filtering_val,
                               FiltersKeys.properties_key.value: {field: filtering_val}})
        return partitions

    @staticmethod
    def _build_partitions_attributes(fields, filtering_values, field_for_label=None):
        """Builds the structure for the list of filtering_values ('partitions') for >1 attributes

        :param fields: Name of the fields concerned
        :param filtering_values: list of values on which do the filtering
        :param field_for_label: If FDU is on several fields, chose one for label initialization.
        :return: The list of filtering values (partitions) created
        """
        partitions = []
        for filtering_vals in filtering_values:
            if len(filtering_vals) != len(fields):
                return None
            if field_for_label is not None and field_for_label not in fields:
                print('Field_for_label "{}" not in fields so ignored'.format(field_for_label))
                field_for_label = None
            element = {FiltersKeys.name_key.value: '', FiltersKeys.properties_key.value: {}}
            for field, filtering_val in zip(fields, filtering_vals):
                if field_for_label is None and not element[FiltersKeys.name_key.value]:
                    element[FiltersKeys.name_key.value] = filtering_val
                elif field_for_label is not None and field == field_for_label:
                    element[FiltersKeys.name_key.value] = filtering_val
                element[FiltersKeys.properties_key.value][field] = filtering_val
            if Filters._label_already_used(partitions, element[FiltersKeys.name_key.value]):
                print('[ERROR FILTERING_VALUES] label', element[FiltersKeys.name_key.value],
                      'already in list')
                continue
            partitions.append(element)
        return partitions

    def _build_list_filtering_values(self, fields, filtering_values, field_for_label=None):
        """Builds the structure for the list of filtering_values (partitions) in a FDU

        :param fields: Fields concerned by the filter.
        :param filtering_values: Filtering values corresponding to the fields for the fdu.
        :param field_for_label: If FDU is on several fields, chose one for label initialization.
        :return: the structure or None if there's a problem in creation
        """
        if len(fields) > 1:
            return self._build_partitions_attributes(fields, filtering_values, field_for_label)
        elif len(fields) == 1:
            return self._build_partition_1_attr(fields[0], filtering_values)
        return None

    def add_FDU_filter(self, name, fields, filtering_values, properties=None, field_for_label=None):
        """Add a new FDU filter to our database

        :param name: Filter name
        :param fields: List of fields concerned by the filter
        :param filtering_values: List of values for the field(s) concerned by the filter
        :param properties: # TODO Determine this parameter
        :param field_for_label: If FDU is on several fields, chose one for label initialization.
        """
        # TODO check if filtering values are in selected field ?
        try:
            list_filtering_values = self._build_list_filtering_values(fields,
                                                                      filtering_values,
                                                                      field_for_label)
            if list_filtering_values is None:
                raise BuildFilterError('[ERROR FDU] Error while creating list of filtering values')
        except BuildFilterError as e:
            print(e.args)
            return
        fdu_filter = FilterFactory.build_filter(name = name, attributes = fields,
                                                filtering_values = list_filtering_values,
                                                properties = properties,
                                                filter_type = "fdu")
        self._add(fdu_filter)

    def add_SQL_filter(self, name, where_clause):
        """Add a new SQL Filter to our database

        :param name: Filter name
        :param where_clause: Values wanted for the filter (eg : 'type = Herbes')
        """
        # Create an instance of the new filter to add
        sql_filter = FilterFactory.build_filter(name = name, where_clause = where_clause,
                                                filter_type = "sql")
        # add it to our list
        self._add(sql_filter)

    def create_filter(self):  # TODO
        """
        :return: a new Filter instance

        ..warning not implemented
        """
        pass

    def _add(self, new_filter):
        """Add a new filter to Orion.

        :param new_filter: The new filter to add.
        """
        if new_filter is None:  # if there was a problem in creation
            return
        self._update()
        # If the filter is already in the db, don't add it
        if self._filter_exist_already(new_filter.name):
            print('Filter "{}" already exist in db'.format(new_filter.name))
            return
        self._prepare_request(new_filter)
        # get the formatted list for request and add this in the DB
        self.request.post(self.url_manager.filter_config_url(),
                          {'value': self._filters_formatted()})
        print('Filter "{}" successfully added !'.format(new_filter.name))

    def add_list(self, new_filters):
        """Add a list of filters to the database.

        :param new_filters: list of all filters and updates (add/delete/edit)
        """
        self._update()
        something_to_add = False  # Avoid to make a request if there's nothing to add
        for filt in new_filters:
            if not self._filter_exist_already(filt.name):
                something_to_add = True
                self._prepare_request(filt)
        if something_to_add:
            self.request.get(self.url_manager.filter_config_url(),
                             {'value': self._filters_formatted()})

    def _prepare_request(self, filt):
        """Add filter to the dictionary (useful for list-modification requests)
        """
        # TODO improve it !!! (work is based on FDUfilter's attributes name...)
        self._filters_as_dic.append(filt.__dict__)

    def _filters_formatted(self):
        """For list-modification request.

        :return: a json formatted string with the same format as required by Orion
        """
        return json.dumps({'dimensions': self._filters_as_dic})

    def _filter_exist_already(self, name, filters=None):
        """Check if a given filter name is already in database

        :rtype: bool
        :param name: the name of the new filter to check
        :param filters: list of filters (up to date with db)
        :return: a boolean to say if filter already added or not
        """
        if filters is None:
            filters = self._elements
        if name in filters:
            print('[ERROR Filter] A filter with the name "{}" already exist in db'.format(name))
            return True
        return False



# https://front.arcopole.fr/orion_fed_105/orion/admin/tree/rightmanagement/groups/__children
from .Subjects import Subjects
from .Group import Group


class Groups(Subjects):
    """
    Class allowing to get access to the defined groups

    """

    def __init__(self):
        super().__init__('groups')

    def _get_system_subject_id(self):
        return ['org', 'public']

    def _create_subject_instance(self, subject_title, subject_id, profile_id):
        return Group(subject_title, subject_id, profile_id)


from .Items import Items
from .Project import Project


class Projects(Items):

    def __init__(self):
        super().__init__("Document Link", "aob_project", lambda data: Project(data))
        self.tags_complement = self.tags+",aob_appId_null"

    def add_item(self, title, text, snippet=None):
        """
        Add a project in aob
        :param title: title of the project
        :param text: content for the project
        :param snippet:
        :return: result of the query (a boolean for "success", the id of the project and the id of the folder)
        """
        url = self.url_manager.add_item_feature_url(self.type)
        data = {"title": title, "snippet": snippet, "text": text, "type": self.type, "tags": self.tags_complement}
        req = self.request.post(url, data).json()
        self._verify(req, "Item \"" + title + "\" was added successfully.\nitemId : " + req["id"],
                     "WARNING : Item \"" + title + "\" was NOT added successfully")
        return req

    def update_item(self, itemId, title, text, snippet=None):
        """
        Update a project in aob
        :param itemId: id of the project you want to update
        :param title: new title of the project
        :param text: content for the project to be updated
        :param snippet:
        :return: result of the query (a boolean for "success" and the id of the project)
        """
        url = self.url_manager.update_item_feature_url(self.type, self._get_owner_from_id(itemId), itemId)
        json_search = self.search(q="id:"+itemId)
        json_url = json_search[0].get("url")
        data = {"title": title, "snippet": snippet, "text": text, "type": self.type, "tags": self.tags_complement,
                "url": json_url}
        req = self.request.post(url, data).json()
        self._verify(req, "Item with id: \"" + itemId + "\" was updated successfully\nTitle : " + title,
                     "WARNING : Item with id: \"" + itemId + "\" was NOT updated successfully")
        return req



# https://front.arcopole.fr/orion_fed_105/orion/admin/tree/rightmanagement/groups/__children
from .Subjects import Subjects
from .User import User


class Users(Subjects):
    """
    Class allowing to get access to the defined users

    """

    def __init__(self):
        super().__init__('users')

    def _get_system_subject_id(self):
        return []

    def _create_subject_instance(self, subject_title, subject_id, profile_id):
        return User(subject_title, subject_id, profile_id)


# ----Apply a filter to a Service----
#  https://front.arcopole.fr/Orion/orion/admin/tree/rightmanagement/profiles/$GROUPID$/aobrights/SERVICES/Nice/Cadastre_MapServer/__configure
# POST
# param : f & token & {"isInherited":false,"rights":[{"filteredDimensions":$filterId$,
#                                                     "action":"write"}]}

# ----Save changes----
# https://front.arcopole.fr/Orion/orion/admin/tree/rightmanagement/__commitPermissions
# POST
# param : f=json & token

# ----Get rights for 1 group on a service----
# https://f106fed.arcopole.fr/Orion/orion/admin/tree/rightmanagement/groups/$GROUPID$/authorizedResources/SERVICES/$SERVICE_ACCESS$
# GET
# f & token

# ----- Two ways of getting rights -----
# ---With groups
# https://f106fed.arcopole.fr/Orion/orion/admin/tree/rightmanagement/groups/$GROUP_ID$/authorizedResources/SERVICES/$LAYER_ACCESS$
# TODO what are start and maxElements attr ?
# f & start & maxElements & token
# ---With profiles
# https://f106fed.arcopole.fr/Orion/orion/admin/tree/rightmanagement/profiles/$GROUP_ID$/aobrights/SERVICES/$LAYER_ACCESS$
# f & start=0 & maxElements=-1 & token
# {
#  "isInherited":false,
#  "rights":
#   [
#      {
#       "action":"read",
#       "filteredDimensions":
#        [
#           "a513c26c-3dd9-49c7-b078-e1dc5043577d",
#           "a89eac34-a739-4fe2-99b0-f94764fac665"
#        ]
#      }
#   ]
# }

from .Layers import Layers
from .Resource import Resource
from .Tables import Tables

from .. import cfg_global


class Service(Resource):
    """Class allowing to handle service resources"""

    def __init__(self, service_id, access_url, capabilities,
                 is_hosted=False, is_managed=False, is_cadastre_resource = False):
        capa = capabilities.split(',')
        super().__init__(access_url = access_url, capabilities = capa)

        self.capabilities = capa
        self._is_managed = is_managed
        self._is_cadastre_resource = is_cadastre_resource
        self.id = service_id
        self.isHosted = is_hosted
        self._layers = Layers(self.access_url)
        self._tables = Tables(self.access_url)

    @property
    def layers(self):
        """
        :return: List of layers corresponding to this service
        """
        return self._layers

    @property
    def tables(self):
        """
        :return: List of tables corresponding to this service
        """
        return self._tables

    def is_shared_with(self, group):
        """Check if the resource is shared with a given group

        :param group: the group to check for
        :return: True if resource shared with group, false otherwise.

        .. warning:: For now, doesn't work"""
        # TODO Keep this method here ?
        # info = self._request_mgr.post_in_python(self._url_builder.resource_sharing_url(self.id))
        # groups = info['sharing']['groups']
        # if group.id in groups:
        #     return True
        # return False
        return True

    # ----- (De)activete sql filter -----

    def activate_sql_filter(self, group, filt):
        """Overwrite this resource method as it is not possible to handle sql filter on a service"""
        print('Impossible to use SQL filters on a service')

    def deactivate_sql_filter(self, group, filt):
        """Overwrite this resource method as it is not possible to handle sql filter on a service"""
        print('Impossible to use SQL filters on a service')

    # ----- Deactivate right-handling from here if service associated to the cadastre resource -----

    def _can_modify(self, group, check_inheritance=True):
        if not super()._can_modify(group, check_inheritance):
            return False
        if self.is_cadastre_resource():
            print("Service associated to cadastre resource. "
                  "Use orion.services.get_cadastre_resource to handle rights for this service")
            return False
        return True

    # ----- Defines management by arcOpole Builder -----
    def enable(self):
        """Enables rights to be managed by arcOpole Builder"""
        if not cfg_global.is_federated:
            print('Method not available if ArcGIS not federated')
            return
        if self.is_managed():
            print('Service {} already enabled, nothing to change'.format(self.get_name()))
            return
        self._change_management(True)

    def disable(self):
        """Disables rights to be managed by arcOpole Builder"""
        if not cfg_global.is_federated:
            print('Method not available if ArcGIS not federated')
            return
        if not self.is_managed():
            print('Service already disabled, nothing to change')
            return
        self._change_management(False)

    def _change_management(self, enable):
        """Change a service's rights management by arcOpole Builder

        :param enable: a boolean to say if want to enable/disable the service
        """
        param = {'f': 'json'}
        if enable:
            param['value'] = '{"managed": true}'
        else:
            param['value'] = '{"managed": false}'
        management_url = self._url_builder.resource_management_url(self.access_url)
        req = self._request_mgr.get_in_python(management_url, param)
        print(req['message'])
        self._is_managed = enable

    # ----- Access methods -----

    @staticmethod
    def get_type():
        return "SERVICE"

    def get_name(self):
        """Override because the 'name' of a Service is its access url"""
        return self.access_url

    def is_cadastre_resource(self):
        """Check if the service is a cadastre_resource or not"""
        return self._is_cadastre_resource

    def __str__(self):
        """Provides a string representation of a service"""
        service_str = ['Service {}; id {};'.format(self.access_url, self.id)]
        if self._is_managed:
            service_str.append(' MANAGED')
        else:
            service_str.append(' NOT MANAGED')
        return "".join(service_str)



# https://front.arcopole.fr/Orion/orion/admin/tree/rightmanagement/profiles/authenticated/aobrights/SERVICES/SampleWorldCities_MapServer/__configure
# {"isInherited":false,"rights":[{"filteredDimensions":["32d0b429-452a-40bd-9865-5a6c6798007a"],"action":"read"}]}

from .Service import Service
from .CadastreResource import CadastreResource
from .. import cfg_global
from ..Elements import Elements


class Services(Elements):
    """
    Class allowing to get access to the list of services
    """

    def __init__(self):
        """Initialize and build our list of services
        """
        super().__init__()
        self._managed_services = {}
        self.handled_service_types = ['FeatureServer', 'MapServer', 'TileServer']

    def urls(self):
        """
        :return: services url as a list
        """
        self._update()
        return list(self._elements.keys())

    # ----- Services handled by aOB -----
    def all_managed(self):
        """Updates and returns list of managed services"""
        if not cfg_global.is_federated:
            print('Method not available if ArcGIS not federated')
            return []
        self._update_managed()
        return self._managed_services.values()

    def get_in_managed(self, element_id):
        """Look for a particular element in managed services.

        :param element_id: identification of element
        :return: required element or None if nothing found
        """
        if not cfg_global.is_federated:
            print('Method not available if ArcGIS not federated')
            return None
        element_id = element_id.strip()
        element = self._update_one_managed(element_id)
        if element is None:
            print("[WARNING] element", element_id,
                  "isn't managed or doesn't exist, None is returned")
            return None
        return element

    def _update_one_managed(self, service_url):
        """Update one service required

        :param service_url: REST access url of the service
        :return: The service if found. None otherwise"""
        list_managed = self.request.get_in_python(self.url_manager.managed_services_url())
        for serv_url in list_managed:
            formatted_url = Services._rreplace(serv_url, '/', '_', 1)
            if formatted_url == service_url:
                return self._create_service(service_url)
        return None

    def _update_managed(self):
        """ Update the list of managed services to make sure it's consistent with our database.
        """
        self._managed_services = {}
        # get list of services managed with orion
        list_managed = self.request.get_in_python(self.url_manager.managed_services_url())
        for serv_url in list_managed:
            formatted_url = Services._rreplace(serv_url, '/', '_', 1)
            self._add_service_to_list(formatted_url, for_managed = True)

    # ----- All services in aOB -----
    def _update(self):
        """ Update the list of services to make sure it's consistent with our database.
        """
        self._elements = {}
        self._build_all()

    def _build_all(self, service_access=""):
        """Recursive method to build a list with all services

        :param service_access: how to access to the service/folder
        """
        node_info_url = self.url_manager.services_children_url(service_access = service_access)
        list_nodes = self.request.get_in_python(node_info_url)
        for node in list_nodes:
            if 'nodeInformationData' in node:
                if service_access:
                    service_url = service_access + "/" + node['name']
                else:
                    service_url = node['name']
                if node['nodeInformationData'] == 'FOLDER':
                    self._build_all(service_url)
                elif node['nodeInformationData'] == 'SERVICE':
                    self._add_service_to_list(service_url)

    def all(self):
        """
        :return: the list of elements' values
        """
        # TODO convert to list ?
        self._update()
        return sorted(self._elements.values(),
                      key = lambda service: service.access_url.lower())

    # ----- Geting 1 service handled by aOB -----
    def get(self, element_name):
        """Look if a particular element exist.

        :param element_name: identification of element
        :return: required element or None if nothing found
        """
        element_name = element_name.strip()
        element = self._update_one(element_name)
        if element is None:
            print("[WARNING] element", element_name, "doesn't exist, None is returned")
            return None
        return element

    def _update_one(self, element_id):
        """Update one service required

        :param element_id: REST access url of the service
        :return: The service if found. None otherwise"""
        return self._create_service(element_id)

    # ----- Getting the cadastre resource ----
    def get_cadastre_resource(self):
        """Get and return the cadastre resource if defined

        :return: the CadastreResource or None if not defined.

        .. deprecated:: AOB 1.4
            Use :func:`orion.businesses.get_cadastre_resource` instead.
        """
        cadastre_req = self._try_to_get_cadastre_resource()

        if cadastre_req is None or not cadastre_req['definition']['serviceReference']:
            return None
        return CadastreResource(cadastre_req)

    def _try_to_get_cadastre_resource(self):
        """Try to get the cadastre resource and handle exception if there is one."""
        cadastre_req = None
        try:
            cadastre_req = self.request.get_in_python(
                self.url_manager.cadastre_definition_url())
        except Exception:
            cadastre_req = None

        return cadastre_req

    # ----- Methods for creation of a service instance -----
    def _add_service_to_list(self, service_url, for_managed=False):
        """Creates an instance of a service and add it to the appropiated list

        :param service_url:
        :param for_managed: parameter saying if fills up common list or list of managed services
        :return:
        """
        service_instance = self._create_service(service_url)
        if service_instance is None:
            return
        if for_managed:
            self._managed_services[service_url] = service_instance
        else:
            self._elements[service_url] = service_instance

    def _create_service(self, service_url):
        """Creates an instance of a service if it is of a good type.

        :param service_url:
        :return: The service created or None if there was an error
        """
        if not self._good_service_type(service_url):
            return None
        # Handle if service ""of good type"" but doesn't exist.
        service_req = self._try_to_get_service(service_url)
        if service_req is None or "error" in service_req:
            # print(service['error'])
            return None
        service = service_req['definition']
        if 'isManaged' in service:  # TODO test on env.is_federated
            is_managed = service['isManaged']
        else:
            is_managed = False
        # Doing so improves retro compatibility for OrionPy
        if 'isCadastreResource' in service:
            is_cadastre_resource = service['isCadastreResource']
        else:
            is_cadastre_resource = False
        return Service(service['portalItemId'], service_url, service['capabilities'],
                       is_managed = is_managed, is_cadastre_resource = is_cadastre_resource)

    def _try_to_get_service(self, service_url):
        """(FIX) Try to get a service and handle exception if there is one."""
        service_req = None
        try:
            service_req = self.request.get_in_python(
                self.url_manager.resource_definition_url(service_url))
        except Exception:
            service_req = None

        return service_req

    def _good_service_type(self, service_name):
        """Look if a given service is of a type that can be managed by aob_admin

        :param service_name: service name. Ends with the service type
        :return: True if the service is of a type handled. False otherwise
        """
        for handled_type in self.handled_service_types:
            if (service_name.lower()).endswith(handled_type.lower()):
                return True
        return False

    @staticmethod
    def _rreplace(original_str, old, new, count):
        """In original_str, replace count occurences of character old by new.
        Starting from the end of the string

        :param original_str:
        :param count:
        :return:
        """
        li = original_str.rsplit(old, count)
        return new.join(li)



from .Resource import Resource, RightLevel


# https://{machine}/{api}/object/SERVICES/{service}/{layerId}/{fieldId}


class Field(Resource):
    """Class allowing to get access and manage a field"""

    # TODO get all corresponding values
    def __init__(self, description, service_url, layer_id, capa, is_managed=False):
        capa = capa.split(',')
        super().__init__(service_url + "/" + str(layer_id) + "/" + str(description['name']),
                         capabilities = capa, parent_access = service_url + "/" + str(layer_id),
                         is_managed = is_managed)
        self.name = description['name']
        self.field_type = description["type"]
        self.alias = description["alias"]
        self.parentUrl = service_url
        self.layerId = layer_id
        self.type = "FIELD"

    # ----- (De)activete filter -----

    def activate_fdu_filter(self, group, filt):
        """Overwrite this resource method as it is not possible to handle filter for a field"""
        print('Impossible to modify filters on a field')

    def deactivate_fdu_filter(self, group, filt):
        """Overwrite this resource method as it is not possible to handle filter for a field"""
        print('Impossible to modify filters on a field')

    def activate_sql_filter(self, group, filt):
        """Overwrite this resource method as it is not possible to handle filter for a field"""
        print('Impossible to modify filters on a field')

    def deactivate_sql_filter(self, group, filt):
        """Overwrite this resource method as it is not possible to handle filter for a field"""
        print('Impossible to modify filters on a field')

    def _can_change_rights(self, group, right_to_apply):
        """Verify if can change the right level for this resource

        :param right_to_apply: New right level to apply
        :param group: id of the group concerned by this modification
        """
        if not super()._can_change_rights(group, right_to_apply):
            return False
        if right_to_apply == RightLevel.WRITE:
            # This case exist only when want to apply editing right
            parent_rights = self._get_parent_rights(group)
            if parent_rights['rights'][0]['action'] != 'write':
                print('To give edit right on a field, containing layer needs to have it as well')
                return False
        return True

    @staticmethod
    def get_type():
        return "FIELD"

    def get_name(self):
        return self.alias

    def __str__(self):
        return "Field {}; alias {}; layer {}; service {}".format(self.name, self.alias,
                                                                 self.layerId, self.parentUrl)



from .orioncsv import OrionCSV
from orionpy.orioncore import cfg_global
from orionpy.orioncore.resources.Table import Table
from orionpy.orioncore.resources.Field import Field
from orionpy.orioncore.resources.Layer import Layer
from orionpy.orioncore.resources.Resource import RightLevel
from orionpy.orioncore.resources.Service import Service


class CSVRightsManagement(OrionCSV):
    def __init__(self, oriongis):
        super().__init__(oriongis)
        self.type_head = 'Type'
        self.url_head = 'Relative url'
        self.label_head = 'Label'
        # TODO also use dictionary value instead of indexes.
        self.type_i = 0
        self.url_i = 1
        self.label_i = 2
        self.inherited_val = 'inherited'

    def generate(self, csv_path, service, group_list=None):
        """Create/Edit a csv file with a summary of all group rights on a service and its components
        (fields, layers, tables, ...)

        :param csv_path: path to the csv files
        :param service: The service for which we want a summary of rights
        :param group_list: (TEMPORARY) groups for which we want rights
        """
        # TODO improve performances !
        if not self._is_service_valid(service):
            return
        print('Generating csv file with a summary of rights for the', service, '...')
        # Builds the header
        header = [self.type_head, self.url_head, self.label_head]

        # If group_list not set, get groups for which service is shared with
        if group_list is None or not group_list:
            group_list = []
            if not cfg_global.is_federated:
                # group_list = self.orion.groups.all()
                print('Not able to get groups if not federated')
                return
            else:
                # TODO how to test if is of OrionGIS type ?
                # if not isinstance(self.orion, OrionGIS) :
                #     print('if you do not use OrionGIS, please specify a list of group !')
                #     return
                groups_ids = self.orion.services_gis_mgr.get_groups_id_shared(service = service)
                for group_id in groups_ids:
                    group_list.append(self.orion.groups.get_with_id(group_id))
        groups = self._add_groups_header(group_list, header)
        if not groups:
            print('None of the group exists..')
            return

        try:
            with open(csv_path, "w+", newline = '') as csv_file:
                writer = self.get_csv_writer(csv_file, header)
                writer.writeheader()
                self._write_resource_information(writer, service, group_list)

                for layer in service.layers.all():
                    print('\tGetting rights for layer "{}"...'.format(layer.get_name()))
                    self._write_resource_information(writer, layer, group_list)
                    for field in layer.fields.all():
                        print('\t\tGetting rights for field "{}"...'.format(field.get_name()))
                        self._write_resource_information(writer, field, group_list)

                for table in service.tables.all():
                    print('\tGetting rights for table "{}"...'.format(table.get_name()))
                    self._write_resource_information(writer, table, group_list)
                    for field in table.fields.all():
                        print('\t\tGetting rights for field "{}"...'.format(field.get_name()))
                        self._write_resource_information(writer, field, group_list)

            print('Done ! Csv file containing rights on resource was generated at', csv_path)
        except IOError as e:
            print(e)

    def _is_service_valid(self, service):
        """Test if can generate a csv of rights on a service

        :param service: The service concerned
        :return: True if there is no problem with the service, False otherwise.
        """
        if service is None:
            print('Service does not exist')
            return False
        if cfg_global.is_federated and not service.is_managed():
            print('Service must be managed by aOB to generate a summary of rights')
            return False
        return True

    def _add_groups_header(self, group_list, header):
        """Adds a grouplist to the header"""
        groups_remaining = False
        for group in group_list:
            if group is not None:
                header.append(group.get_name())
                groups_remaining = True
        if not groups_remaining:
            print('No group to get')
            return False
        return group_list

    def _write_resource_information(self, writer, resource, group_list):
        """Writes one resource rights

        :param writer: the csv writer
        :param resource: resource concerned
        :param group_list: groups for which write information
        """
        row_dict = {self.type_head: resource.get_type(),
                    self.url_head: resource.get_access_url(),
                    self.label_head: resource.get_name()}
        for group in group_list:
            if group is not None:
                right = resource.get_resolved_right(group).value

                if resource.has_inherited_right(group):
                    row_dict[group.get_name()] = "{} ({})".format(self.inherited_val, right)
                else:
                    row_dict[group.get_name()] = right
        writer.writerow(row_dict)

    def read_and_apply(self, csv_path, handle_sharing=False):
        """Parse a csv file with resources, groups and rights.
        Does so in order to update rights for groups

        :param csv_path: path to the csv files
        :param handle_sharing: Boolean saying if handle sharing with groups or not
        """
        print('Scanning file', csv_path, '....')
        dic_resource = {Service.get_type(): None,
                        Layer.get_type(): None,
                        Table.get_type(): None}
        try:
            with open(csv_path, "r") as csv_file:
                # TODO use DictReader ?
                reader = self.get_csv_reader(csv_file)
                # Prepares data
                groups = self._get_groups(header = next(reader), start = 3)
                if not groups:
                    print('None of the group exists..')
                    return

                sharing_checked = False
                # Get data for each row and, if possible, update right
                for row in reader:
                    resource = self._get_correct_resource(row, dic_resource)
                    if handle_sharing and not sharing_checked and \
                            dic_resource[Service.get_type()] is not None:
                        self._handle_sharing(dic_resource[Service.get_type()], groups)
                        sharing_checked = True  # avoid calling this method a million time
                    if resource is None:
                        continue
                    # TODO reduce number of calls to update if not needed.
                    self._update_resource_right(resource, row, groups)
            print('Done ! All possible rights values were successfully updated !')
        except IOError as e:
            print(e)

    def _handle_sharing(self, service, groups):
        """ If service not shared w/ group, share it with it"""
        for group in groups:
            if group is not None and \
                    not self.orion.services_gis_mgr.is_shared_with(service, group.get_id()):
                self.orion.services_gis_mgr.share(service, group.get_id())

    def _get_correct_resource(self, row, resource_d):
        """Based on type in row and resources got before, return correct resource

        :return: Resource or None if problem getting resource"""
        if row[self.type_i] == Service.get_type():
            resource_d[Service.get_type()] = self.orion.services.get_in_managed(row[self.url_i])
            return resource_d[Service.get_type()]
        elif row[self.type_i] == Layer.get_type() and resource_d[Service.get_type()] is not None:
            # Get layer's id from REST url.
            layer_id = row[self.url_i].split('/')[-1]
            resource_d[Layer.get_type()] = resource_d[Service.get_type()].layers.get_with_id(
                layer_id)
            return resource_d[Layer.get_type()]
        elif row[self.type_i] == Table.get_type() and resource_d[Service.get_type()] is not None:
            resource_d[Layer.get_type()] = None  # for field's right handling
            table_id = row[self.url_i].split('/')[-1]
            resource_d[Table.get_type()] = resource_d[Service.get_type()].tables.get_with_id(
                table_id)
            return resource_d[Table.get_type()]
        elif row[self.type_i] == Field.get_type():
            # Get field's name from REST url.
            field_name = row[self.url_i].split('/')[-1]
            # NB : no need to save field in this case
            if resource_d[Layer.get_type()] is not None:
                return resource_d[Layer.get_type()].fields.get(field_name)
            elif resource_d[Table.get_type()] is not None:
                return resource_d[Table.get_type()].fields.get(field_name)
        print('Unknown type', row[self.type_i])
        return None

    def _update_resource_right(self, resource, row, groups):
        """Check and update a resource's right for each group of groups."""
        if resource is None:
            print('Resource', row[1:3], 'unknown')
            return
        for val, group in zip(row[3:], groups):
            if group is None:  # The group doesn't exist
                continue
            self._update_good_right(val, group, resource)

    def _update_good_right(self, val, group, resource):
        """Calls the good update_right method based on val"""
        val = val.strip()  # handle trailing blanklines
        right = None
        # If empty string read, treated the same way as 'inherited'
        if val.startswith(self.inherited_val) or not val:
            resource.enable_inheritance(group)
            return

        # TODO "pythonic switch" ? Or get_enum_from_value method ?
        if val == RightLevel.ACCESS.value:
            right = RightLevel.ACCESS
        elif val == RightLevel.READ.value:
            right = RightLevel.READ
        elif val == RightLevel.WRITE.value:
            right = RightLevel.WRITE
        else:
            print('ERROR : wrong right type "', val, '" in csv')
            return
        resource.force_right(group, right)



# https://front.arcopole.fr/Orion/orion/admin/tree/rightmanagement/profiles/authenticated/aobrights/SERVICES/SampleWorldCities_MapServer/__configure
# {"isInherited":false,"rights":[{"filteredDimensions":["32d0b429-452a-40bd-9865-5a6c6798007a"],"action":"read"}]}

# ---- Get layers corresponding to a service :
# https://front1051.arcopole.fr/Orion/orion/admin/tree/object/SERVICES/Cannes/EspacesVerts_FeatureServer
# f & token
# "definition":
# {
#     "layers":
#         [
#             {
#                 "id":$LAYER_ID$,
# "name":$LAYER_NAME$,
# "parentLayerId":$parentLayerId$,
# "defaultVisibility": bool,
# "subLayerIds": null,
# "minScale": 0,
# "maxScale": 0
# },
# ],

from ..Elements import Elements
from .Layer import Layer


class Layers(Elements):
    """
    Class allowing to get access to a list of Layers for a particular service
    """

    def __init__(self, parent_service_url):
        """Initialize and build our list of layers

        :param parent_service_url: url (unique) for the service containing all those layers
        """
        super().__init__()
        self.service_url = parent_service_url

    def _update(self):
        """Update list to be consistent with database
        """
        self._elements = {}
        service_def_url = self.url_manager.resource_definition_url(self.service_url)
        service_def = self.request.get_in_python(service_def_url)
        for layer in service_def['definition']['layers']:
            if 'isManaged' in service_def['definition']:
                is_managed = service_def['definition']['isManaged']
            else:
                is_managed = False
            self._elements[layer['id']] = Layer(layer, self.service_url,
                                                service_def['definition']['capabilities'],
                                                is_managed)

    def get_id(self, layer_name):
        """With the name of a layer, allows to get its id

        :param layer_name: Name of layer for which we want the id
        :return: Layer id or None if nothing was found
        """
        layer = self.get(layer_name)
        if layer is not None:
            return layer.id
        return None

    def get_with_id(self, layer_id):
        """Returns a layer given its id
        """
        self._update()
        # Expects an int so there won't be any space.
        if isinstance(layer_id, str):
            layer_id = layer_id.strip()
            layer_id = int(layer_id)
        if layer_id in self._elements:
            return self._elements[layer_id]
        print("[WARNING] id", layer_id, "doesn't exist, None is returned")
        return None

    def get(self, element_name):
        """Look for a particular layer in elements

        :param element_name: name of layer
        :return: First Layer with this idor None if nothing was found"""
        self._update()
        element_name = element_name.strip()
        for key, layer in self._elements.items():
            if layer.name == element_name:
                return layer
        print("[WARNING] layer", element_name, "doesn't exist, None is returned")
        return None

    def all(self):
        """
        :return: the list of elements' values
        """
        self._update()
        return list(self._elements.values())



import csv


class OrionCSV:
    # TODO factorize implementation of read_apply, generate
    def __init__(self, orion):
        self.orion = orion
        self.delimiter = ';'
        self.quoting = csv.QUOTE_MINIMAL
        self.csv_file = None

    def get_csv_writer(self, csv_file, header):
        """Get a csv writer from a given file

        :param header: header of file
        :param csv_file: path to a csv file
        :return: a csv dictionary writer for the given csv file
        """
        return csv.DictWriter(csv_file, fieldnames = header,
                              delimiter = self.delimiter, quoting = self.quoting)

    def get_csv_reader(self, csv_file):
        """Get a csv reader from a given file

        :param csv_file: path to a csv file
        :return: a reader object for the given csv file
        """
        # TODO return a DictReader instead ?
        return csv.reader(csv_file, delimiter = self.delimiter, quoting = self.quoting)

    def _get_groups(self, header, start):
        """Builds list of groups based on csv header"""
        groups = []
        nothing_added = True
        for group_name in header[start:]:
            group = self.orion.groups.get(group_name)
            if group is None:
                print('Group "', group_name, '" does not exist')
            else:
                nothing_added = False
            # To avoid problems later, add value in every cases
            groups.append(group)
        if nothing_added:
            return False
        else:
            return groups



# similarities with requirements.txt
# Gestion de la distribution et de la creation de paquets
from setuptools import setup, find_packages
from os import path

# Gets current path
here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding = 'utf-8') as f:
    long_description = f.read()

# Gets the version to set from the VERSION file
# Version meets the following : https://www.python.org/dev/peps/pep-0440/#version-scheme
with open(path.join(here, 'VERSION')) as version_file:
    version = version_file.read().strip()

setup(
    name = 'orionpy',  # Required
    version = version,  # content of VERSION file
    description = 'Python API to administrate arcOpole Builder',
    long_description = long_description,  # content of README.md file
    long_description_content_type = "text/markdown",
    url = 'https://gitlab.com/esrifrance-public/orionpy/orionpy',
    author = 'acollange',
    author_email = 'acollange@esrifrance.fr',
    license="Proprietary",
    # Keeps docs and orionpy. Excludes orion[core,gis,csv] to be importable packages
    packages = find_packages(exclude = ['demo', 'tests', 'test', 'orioncore', 'oriongis', 'orioncsv']),
    install_requires = [
        'requests',
        'urllib3',
    ],
    # Classifiers gives some more information about the project
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3 :: Only',
        'License :: Other/Proprietary License'
    ],
    python_requires='>=3.6',
    # Keywords related to this project
    keywords = 'orionpy orion arcopole builder',
)




from orionpy.orioncore.resources.Service import Service


class ServiceGIS(Service):
    def __init__(self, service_id, access_url, capabilities, is_hosted=False, is_managed=True):
        super().__init__(service_id, access_url, capabilities, is_hosted, is_managed)
